#
# Copyright 2019-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""User command line interface of Numina, verify functionality."""

import logging
import os

from .helpers import create_datamanager, load_observations
from numina.util.context import working_directory


_logger = logging.getLogger(__name__)


def register(subparsers, config):
    parser_verify = subparsers.add_parser(
        'verify',
        help='verify a observation result'
    )

    parser_verify.set_defaults(command=verify)

    parser_verify.add_argument(
        '-c', '--task-control', dest='reqs',
        help='configuration file of the processing task', metavar='FILE'
    )
    parser_verify.add_argument(
        '-r', '--requirements', dest='reqs',
        help='alias for --task-control', metavar='FILE'
    )
    parser_verify.add_argument(
        '-i', '--instrument', dest='insconf',
        default=None,
        help='name of an instrument configuration'
    )
    parser_verify.add_argument(
        '--profile-path', dest='profilepath',
        default=None,
        help='location of the instrument profiles'
    )
    parser_verify.add_argument(
        '-p', '--pipeline', dest='pipe_name',
        default='default', help='name of a pipeline'
    )
    parser_verify.add_argument(
        '--basedir', action="store", dest="basedir",
        help='path to create the following directories'
    )
    parser_verify.add_argument(
        '--datadir', action="store", dest="datadir",
        help='path to directory containing pristine data'
    )
    parser_verify.add_argument(
        '--resultsdir', action="store", dest="resultsdir",
        help='path to directory to store results'
    )
    parser_verify.add_argument(
        '--workdir', action="store", dest="workdir",
        help='path to directory containing intermediate files'
    )
    parser_verify.add_argument(
        '--cleanup', action="store_true", dest="cleanup",
        default=False, help='cleanup workdir on exit [disabled]'
    )
    parser_verify.add_argument(
        '--not-copy-files', action="store_const", dest="copy_files", const=False,
        help='do not copy observation result and requirement files'
    )
    parser_verify.add_argument(
        '--link-files', action="store_const", dest="copy_files", const=False,
        help='do not copy observation result and requirement files'
    )
    parser_verify.add_argument(
        '--dump-control', action="store_true",
        help='save the modified task control file'
    )
    parser_verify.add_argument(
        '--session', action="store_true",
        help='use the obresult file as a session file'
    )
    parser_verify.add_argument(
        '--validate', action="store_const", const=True,
        help='validate inputs and results of recipes'
    )
    parser_verify.add_argument(
        'files', nargs='+',
        help='file with the observation result'
    )

    return parser_verify


def verify(args, extra_args, config):

    # Loading observation result if exists
    sessions, loaded_obs = load_observations(args.obsresult, args.session)

    # Override like this
    if args.basedir:
        config['tool.run']['basedir'] = args.basedir
    if args.datadir:
        config['tool.run']['datadir'] = args.datadir

    datamanager = create_datamanager(config, args.reqs, extra_args.extra_control)
    datamanager.backend.add_obs(loaded_obs)

    # Start processing
    jobs = []
    for session in sessions:
        for job in session:
            if job['enabled']:
                jobs.append(job)

    for job in jobs:
        run_verify(
            datamanager, job['id'], copy_files=args.copy_files,
            validate_inputs=args.validate, validate_results=args.validate
        )
    if args.obs:
        # verify as oblocks
        # Loading observation result if exists
        sessions, loaded_obs = load_observations(args.files, args.session)
        datamanager = create_datamanager(config, None, args.basedir, args.datadir)
        datamanager.backend.add_obs(loaded_obs)

        # Start processing
        jobs = []
        for session in sessions:
            for job in session:
                if job['enabled']:
                    jobs.append(job)

        for job in jobs:
            run_verify(
                datamanager, job['id'], copy_files=args.copy_files,
                validate_inputs=True, validate_results=True
            )
    else:
        # This function loads the recipes
        datamanager = create_datamanager(config, None, args.basedir, args.datadir)
        for file in args.files:
            _logger.info(f'checking {file}')
            try:
                result = check_file(file)
            except Exception:
                result = False
                # _logger.warning('with error {}'.format(error))
            _logger.info(f'checked {file}, valid={result}')
            # print('done')

    return 0


def run_verify(datastore, obsid, as_mode=None, requirements=None, copy_files=False,
               validate_inputs=False, validate_results=False):
    """Verify raw images"""

    configuration = 'default'
    _logger.info("verify OB with id={}".format(obsid))
    _logger.info(f"verify OB with id={obsid}")

    # Roll back to cwd after leaving the context
    with working_directory(datastore.datadir):

        obsres = datastore.backend.obsres_from_oblock_id(
            obsid, as_mode=as_mode, configuration=configuration
        )

        thisdrp = datastore.backend.drps.query_by_name(obsres.instrument)

        msg = f'the mode of this obsres is {obsres.instrument}.{obsres.mode}'
        _logger.info(msg)

        for v in thisdrp.modes.values():
            if v.key == obsres.mode:
                mode_obj = v
                break
        else:
            raise ValueError(f'unrecognized mode {obsres.mode}')

        image_is = mode_obj.rawimage

        for f in obsres.frames:
            with f.open() as hdulist:
                _logger.debug(f'checking {f.filename}')
                try:
                    check_image(hdulist, astype=image_is)
                except Exception:
                    pass
        _logger.info('Checking that individual images are valid for this mode')
        mode_obj.validate(obsres)
        #


def check_file(filename, astype=None, level=None):
    import json
    import yaml
    import astropy.io.fits as fits

    json_ext = ['.json']
    fits_ext = ['.fits', '.fits.gz', '.fit']
    yaml_ext = ['.yaml', '.yml']

    fname, ext = os.path.splitext(filename)

    if ext in json_ext:
        # print('as json')
        with open(filename) as fd:
            obj = json.load(fd)
        return check_json(obj, astype=astype, level=level)
    elif ext in fits_ext:
        # print('as fits')
        with fits.open(filename) as img:
            return check_image(img, astype=None)
    elif ext in yaml_ext:
        with open(filename) as fd:
            obj = list(yaml.safe_load_all(fd))
        return check_yaml(obj, astype=astype, level=level)
    else:
        print(f'ignoring {filename}')
        return True


def check_json(obj, astype=None, level=None):
    import numina.core.config as cfg

    if 'instrument' in obj:
        instrument = obj['instrument']
        return cfg.check(instrument, obj)
    else:
        print("no 'instrument' field in object")
        # Try to check with other schema
        return False


def check_yaml(obj, astype=None, level=None):
    # import numina.core.config as cfg

    # import jsonschema
    # import json
    # try to verify as obsblock
    # path = "/home/spr/devel/guaix/numina/schemas/oblock-schema.json"
    # with open(path) as fd:
    #     schema = json.load(fd)
    #     jsonschema.validate(obj, schema=schema)
    pass


def check_image(hdulist, astype=None, level=None):

    import numina.core.config as cfg
    # Determine the instrument name
    hdr = hdulist[0].header
    instrument = hdr['INSTRUME']
    return cfg.check(instrument, hdulist, astype=astype, level=level)


def convert_headers(hdulist):
    headers = [convert_header(hdu.header) for hdu in hdulist]
    return headers


def convert_header(header):
    hdu_v = {}
    hdu_c = {}
    hdu_o = []
    hdu_repr = {'values': hdu_v, 'comments': hdu_c, 'ordering': hdu_o}

    for card in header.cards:
        key = card.keyword
        value = card.value
        comment = card.comment
        hdu_v[key] = value
        hdu_c[key] = comment
        hdu_o.append(key)

    return hdu_repr



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Unit test for types"""

from ..datatype import NullType, DataType


def test_null_type():
    """Test NullType."""

    nullt = NullType()

    values = [None, 1, 1.0, [1, 2, 3], {'a': 1, 'b': 2}]

    for val in values:
        assert nullt.convert(val) is None

    for val in values:
        assert nullt.validate(nullt.convert(val))


def test_convert_in_out():

    class AVal(object):
        pass

    class A(DataType):

        def convert(self, obj):
            return 1

    a = A(ptype=AVal)
    assert a.convert_in(None) == 1
    assert a.convert_out(None) == 1

    class B(A):

        def convert_out(self, obj):
            return 2

    b = B(ptype=AVal)
    assert b.convert_in(None) == 1
    assert b.convert_out(None) == 2



#
# Copyright 2008-2015 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Recipes for system checks. """

import logging

from numina.core import BaseRecipe, DataFrameType
from numina.core.requirements import ObservationResultRequirement
import numina.core.dataholders as dh
import numina.core.query as qry


_logger = logging.getLogger(__name__)


class AlwaysFailRecipe(BaseRecipe):
    """A Recipe that always fails."""

    def __init__(self, *args, **kwargs):
        super(AlwaysFailRecipe, self).__init__(
            version="1"
        )

    def run(self, requirements):
        raise TypeError('This Recipe always fails')


class AlwaysSuccessRecipe(BaseRecipe):
    """A Recipe that always successes."""

    def __init__(self, *args, **kwargs):
        super(AlwaysSuccessRecipe, self).__init__(
            version=1
        )

    def run(self, recipe_input):
        return self.create_result()


class OBSuccessRecipe(BaseRecipe):
    """A Recipe that always successes, it requires an OB"""

    obresult = ObservationResultRequirement()

    def __init__(self, *args, **kwargs):
        super(OBSuccessRecipe, self).__init__(
            version=1
        )

    def run(self, recipe_input):
        return self.create_result()


class Combine(BaseRecipe):

    obresult = ObservationResultRequirement()
    method = dh.Parameter('mean', "Method of combination")
    method_kwargs = dh.Parameter(
        {}, "Arguments passed to the combination method")
    field = dh.Parameter('image', "Extract field of previous result")
    result = dh.Result(DataFrameType)

    def run(self, recipe_input):
        import numina.array.combine as c
        from numina.processing.combine import combine_frames

        method = getattr(c, recipe_input.method)
        obresult = recipe_input.obresult
        method_kwargs = recipe_input.method_kwargs
        result = combine_frames(obresult.frames,
                                method, method_kwargs=method_kwargs
                                )
        return self.create_result(result=result)

    def build_recipe_input(self, obsres, dal):
        import numina.exceptions

        result = {}
        result['obresult'] = obsres
        for key in ['method', 'field']:
            req = self.requirements()[key]
            query_option = self.query_options.get(key)
            try:
                result[key] = req.query(dal, obsres, options=query_option)
            except numina.exceptions.NoResultFound as notfound:
                req.on_query_not_found(notfound)
        if 'field' in result:
            # extract values:
            obreq = self.requirements()['obresult']
            qoptions = qry.ResultOf(field=result['field'])

            obsres = obreq.query(dal, obsres, options=qoptions)
            result['obresult'] = obsres

        rinput = self.create_input(**result)
        return rinput



import numpy

from numina.array.peaks.detrend import detrend

numpy.random.seed(12812818)


def test_compute_trend():
    numpy.random.seed(12812818)

    xx = numpy.arange(100)
    yy = 700 + 3.2 * (xx / 100.0)
    yy = numpy.random.normal(loc=yy, scale=2.0)

    trend = detrend(yy)

    maxt = trend.max()
    mint = trend.min()
    expected = [704.812599, 698.595538331]

    numpy.testing.assert_array_almost_equal([maxt, mint], expected)



#
# Copyright 2008-2022 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import logging
import itertools

import numpy
import scipy.stats
import scipy.ndimage
from numina.array.blocks import max_blk_coverage, blk_nd_short


# Values stored in integer masks
PIXEL_HOT = 1
PIXEL_DEAD = 1
PIXEL_VALID = 0

#
HIGH_SIGMA = 200
LOW_SIGMA = -200

_logger = logging.getLogger(__name__)


def update_mask(mask, gmask, newmask, value):
    f1_mask = mask[gmask]
    f1_mask[newmask] = value
    mask[gmask] = f1_mask
    gmask = mask == PIXEL_VALID
    smask = mask != PIXEL_VALID
    return mask, gmask, smask


# IRAF task
def ccdmask(flat1, flat2=None, mask=None, lowercut=6.0, uppercut=6.0,
            siglev=1.0, mode='region', nmed=(7, 7), nsig=(15, 15)):
    """Find cosmetic defects in a detector using two flat field images.

    Two arrays representing flat fields of different exposure times are
    required. Cosmetic defects are selected as points that deviate
    significantly of the expected normal distribution of pixels in
    the ratio between `flat2` and `flat1`. The median of the ratio
    is computed and subtracted. Then, the standard deviation is estimated
    computing the percentiles
    nearest to the pixel values corresponding to`siglev` in the normal CDF.
    The standard deviation is then the distance between the pixel values
    divided by two times `siglev`. The ratio image is then normalized with
    this standard deviation.

    The behavior of the function depends on the value of the parameter
    `mode`. If the value is 'region' (the default), both the median
    and the sigma are computed in boxes. If the value is 'full', these
    values are computed using the full array.

    The size of the boxes in 'region' mode is given by `nmed` for
    the median computation and `nsig` for the standard deviation.

    The values in the normalized ratio array above `uppercut`
    are flagged as hot pixels, and those below '-lowercut` are
    flagged as dead pixels in the output mask.

    :parameter flat1: an array representing a flat illuminated exposure.
    :parameter flat2: an array representing a flat illuminated exposure.
    :parameter mask: an integer array representing initial mask.
    :parameter lowercut: values below this sigma level are flagged as dead pixels.
    :parameter uppercut: values above this sigma level are flagged as hot pixels.
    :parameter siglev: level to estimate the standard deviation.
    :parameter mode: either 'full' or 'region'
    :parameter nmed: region used to compute the median
    :parameter nsig: region used to estimate the standard deviation
    :returns: the normalized ratio of the flats, the updated mask and standard deviation

    .. note::

        This function is based on the description of the task
        ccdmask of IRAF

    .. seealso::

        :py:func:`cosmetics`
            Operates much like this function but computes
            median and sigma in the whole image instead of in boxes

    """

    if flat2 is None:
        # we have to swap flat1 and flat2, and
        # make flat1 an array of 1s
        flat1, flat2 = flat2, flat1
        flat1 = numpy.ones_like(flat2)

    if mask is None:
        mask = numpy.zeros_like(flat1, dtype='int')

    ratio = numpy.zeros_like(flat1)
    invalid = numpy.zeros_like(flat1)
    invalid[mask == PIXEL_HOT] = HIGH_SIGMA
    invalid[mask == PIXEL_DEAD] = LOW_SIGMA

    gmask = mask == PIXEL_VALID
    _logger.info('valid points in input mask %d', numpy.count_nonzero(gmask))
    smask = mask != PIXEL_VALID
    _logger.info('invalid points in input mask %d', numpy.count_nonzero(smask))

    # check if there are zeros in flat1 and flat2
    zero_mask = numpy.logical_or(flat1[gmask] <= 0, flat2[gmask] <= 0)

    # if there is something in zero mask
    # we update the mask
    if numpy.any(zero_mask):
        mask, gmask, smask = update_mask(mask, gmask, zero_mask, PIXEL_DEAD)
        invalid[mask == PIXEL_DEAD] = LOW_SIGMA

    # ratio of flats
    ratio[gmask] = flat2[gmask] / flat1[gmask]
    ratio[smask] = invalid[smask]

    if mode == 'region':
        _logger.info('computing median in boxes of %r', nmed)
        ratio_med = scipy.ndimage.median_filter(ratio, size=nmed)
        # subtracting the median map
        ratio[gmask] -= ratio_med[gmask]
    else:
        _logger.info('computing median in full array')
        ratio_med = numpy.median(ratio[gmask])
        ratio[gmask] -= ratio_med

    # Quantiles that contain nsig sigma in normal distribution
    qns = 100 * scipy.stats.norm.cdf(siglev)
    pns = 100 - qns
    _logger.info('percentiles at siglev=%f', siglev)
    _logger.info('low %f%% high %f%%', pns, qns)

    # in several blocks of shape nsig
    # we estimate sigma
    sigma = numpy.zeros_like(ratio)

    if mode == 'region':
        mshape = max_blk_coverage(blk=nsig, shape=ratio.shape)
        _logger.info('estimating sigma in boxes of %r', nsig)
        _logger.info('shape covered by boxes is  %r', mshape)
        block_gen = blk_nd_short(blk=nsig, shape=ratio.shape)
    else:
        mshape = ratio.shape
        _logger.info('estimating sigma in full array')
        # slice(None) is equivalent to [:]
        block_gen = itertools.repeat(slice(None), 1)

    for blk in block_gen:
        # mask for this region
        m = mask[blk] == PIXEL_VALID
        valid_points = numpy.ravel(ratio[blk][m])
        ls = scipy.stats.scoreatpercentile(valid_points, pns)
        hs = scipy.stats.scoreatpercentile(valid_points, qns)

        _logger.debug('score at percentiles')
        _logger.debug('low %f high %f', ls, hs)

        # sigma estimation
        sig = (hs - ls) / (2 * siglev)
        _logger.debug('sigma estimation is %f ', sig)

        # normalized points
        sigma[blk] = sig

    # fill regions of sigma not computed
    fill0 = ratio.shape[0] - mshape[0]
    fill1 = ratio.shape[1] - mshape[1]
    if fill0 > 0:
        _logger.info('filling %d rows in sigma image', fill0)
        sigma[:, mshape[0]:] = sigma[:, mshape[0] - fill0:mshape[0]]

    if fill1 > 0:
        _logger.info('filling %d columns in sigma image', fill1)
        sigma[mshape[1]:, :] = sigma[mshape[1] - fill1:mshape[1], :]

    # invalid_sigma = sigma <= 0.0
    # if numpy.any(invalid_sigma):
    #     _logger.info('updating mask with points where sigma <=0')
    #     mask, gmask, smask = update_mask(mask, gmask, invalid_sigma, PIXEL_HOT)
    #     invalid[mask == PIXEL_HOT] = HIGH_SIGMA

    ratio[gmask] /= sigma[gmask]

    f1_ratio = ratio[gmask]
    f1_mask = mask[gmask]
    f1_mask[f1_ratio >= uppercut] = PIXEL_HOT
    f1_mask[f1_ratio <= -lowercut] = PIXEL_DEAD
    mask[gmask] = f1_mask

    return ratio, mask, sigma


def robust_std(valid, central, siglev):
    import scipy.stats
    qns = 100 * scipy.stats.norm.cdf(siglev)
    pns = 100 - qns

    ls = scipy.stats.scoreatpercentile(valid.flat - central, pns)
    hs = scipy.stats.scoreatpercentile(valid.flat - central, qns)

    # sigma estimation
    sig = (hs - ls) / (2 * siglev)

    return sig


def comp_ratio(img1, img2, mask):
    mask = mask == 1
    mask1 = img1 <= 0
    mask2 = img2 <= 0
    mask3 = mask1 | mask2 | mask
    with numpy.errstate(divide='ignore', invalid='ignore'):
        ratio = img1 / img2
    ratio[mask3] = 0.0
    return ratio, mask3


def cosmetics(flat1, flat2=None, mask=None, lowercut=6.0, uppercut=6.0, siglev=2.0):
    """Find cosmetic defects in a detector using two flat field images.

    Two arrays representing flat fields of different exposure times are
    required. Cosmetic defects are selected as points that deviate
    significantly of the expected normal distribution of pixels in
    the ratio between `flat2` and `flat1`.

    The median of the ratio array is computed and subtracted to it.

    The standard deviation of the distribution of pixels is computed
    obtaining the percentiles nearest the pixel values corresponding to
    `nsig` in the normal CDF. The standar deviation is then the distance
    between the pixel values divided by two times `nsig`.
    The ratio image is then normalized with this standard deviation.

    The values in the ratio above `uppercut` are flagged as hot pixels,
    and those below `-lowercut` are flagged as dead pixels in the output mask.

    :parameter flat1: an array representing a flat illuminated exposure.
    :parameter flat2: an array representing a flat illuminated exposure.
    :parameter mask: an integer array representing initial mask.
    :parameter lowercut: values bellow this sigma level are flagged as dead pixels.
    :parameter uppercut: values above this sigma level are flagged as hot pixels.
    :parameter siglev: level to estimate the standard deviation.
    :returns: the updated mask

    """

    if flat2 is None:
        flat1, flat2 = flat2, flat1
        flat1 = numpy.ones_like(flat2)

    if type(mask) is not numpy.ndarray:
        mask = numpy.zeros(flat1.shape, dtype='int')

    ratio, mask = comp_ratio(flat1, flat2, mask)
    fratio1 = ratio[~mask]
    central = numpy.median(fratio1)
    std = robust_std(fratio1, central, siglev)
    mask_u = ratio > central + uppercut * std
    mask_d = ratio < central - lowercut * std
    mask_final = mask_u | mask_d | mask
    return mask_final


if __name__ == "__main__":
    flat1 = numpy.zeros((2, 2))
    flat1[0] = 1
    flat2 = flat1
    mascara = cosmetics(flat1, flat2)
    print(mascara)
    print(mascara.sum())



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import sys

from numina.array.display.matplotlib_qt import plt

DEBUGPLOT_CODES = (0, -1, 1, -2, 2, -10, 10, -11, 11, -12, 12,
                   -21, 21, -22, 22)


def pause_debugplot(debugplot, optional_prompt=None, pltshow=False,
                    tight_layout=True):
    """Ask the user to press RETURN to continue after plotting.

    Parameters
    ----------
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed:
        00 : no debug, no plots
        01 : no debug, plots without pauses
        02 : no debug, plots with pauses
        10 : debug, no plots
        11 : debug, plots without pauses
        12 : debug, plots with pauses
        21 : debug, extra plots without pauses
        22 : debug, extra plots with pause
        NOTE: negative values are also valid and indicate that a call
        to plt.close() is also performed
    optional_prompt : string
        Optional prompt.
    pltshow : bool
        If True, a call to plt.show() is also performed.
    tight_layout : bool
        If True, and pltshow=True, a call to plt.tight_layout() is
        also performed.

    """

    if debugplot not in DEBUGPLOT_CODES:
        raise ValueError('Invalid debugplot value:', debugplot)

    if debugplot < 0:
        debugplot_ = -debugplot
        pltclose = True
    else:
        debugplot_ = debugplot
        pltclose = False

    if pltshow:
        if debugplot_ in [1, 2, 11, 12, 21, 22]:
            if tight_layout:
                plt.tight_layout()
            if debugplot_ in [1, 11, 21]:
                plt.show(block=False)
                plt.pause(0.2)
            elif debugplot_ in [2, 12, 22]:
                print('Press "q" to continue...', end='')
                sys.stdout.flush()
                plt.show()
                print('')
    else:
        if debugplot_ in [2, 12, 22]:
            if optional_prompt is None:
                print('Press <RETURN> to continue...', end='')
            else:
                print(optional_prompt, end='')
            sys.stdout.flush()
            cdummy = sys.stdin.readline().strip()  # noqa: F841

    if debugplot_ in [1, 2, 11, 12, 21, 22] and pltclose:
        plt.close()



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import numpy
from scipy.interpolate import interp1d
import scipy.ndimage as ndimage

from .blocks import blockgen1d, blockgen  # noqa: F401
from .imsurfit import FitOne


def subarray_match(shape, ref, sshape, sref=None):
    """Compute the slice representation of intersection of two arrays.

    Given the shapes of two arrays and a reference point ref, compute the
    intersection of the two arrays.
    It returns a tuple of slices, that can be passed to the two \
    images as indexes

    :param shape: the shape of the reference array
    :param ref: coordinates of the reference point in the first array system
    :param sshape: the shape of the second array
    :param: sref: coordinates of the reference point in the \
    second array system, the origin by default
    :type sref: sequence or None
    :return: two matching slices, corresponding to both arrays \
    or a tuple of Nones if they don't match
    :rtype: a tuple

    Example:

      >>> import numpy
      >>> im = numpy.zeros((1000, 1000))
      >>> sim = numpy.ones((40, 40))
      >>> i,j = subarray_match(im.shape, [20, 23], sim.shape)
      >>> im[i] = 2 * sim[j]

    """
    # Reference point in im
    ref1 = numpy.asarray(ref, dtype='int')

    if sref is not None:
        ref2 = numpy.asarray(sref, dtype='int')
    else:
        ref2 = numpy.zeros_like(ref1)

    offset = ref1 - ref2
    urc1 = numpy.minimum(offset + numpy.asarray(sshape) -
                         1, numpy.asarray(shape) - 1)
    blc1 = numpy.maximum(offset, 0)
    urc2 = urc1 - offset
    blc2 = blc1 - offset

    def valid_slice(b, u):
        if b >= u + 1:
            return None
        else:
            return slice(b, u + 1)

    f = tuple(valid_slice(b, u) for b, u in zip(blc1, urc1))
    s = tuple(valid_slice(b, u) for b, u in zip(blc2, urc2))

    if not all(f) or not all(s):
        return None, None

    return f, s


def combine_shape(shapes, offsets):
    # Computing final array size and new offsets
    sharr = numpy.asarray(shapes)
    offarr = numpy.asarray(offsets)
    ucorners = offarr + sharr
    ref = offarr.min(axis=0)
    finalshape = tuple(ucorners.max(axis=0) - ref)
    offsetsp = offarr - ref
    return finalshape, offsetsp


def combine_shapes(shapes, refs, order='rc'):

    # Order 'rc' -> row column
    # Order 'xy' -> xy coordinates

    sharr = numpy.asarray(shapes)
    rfarr = numpy.asarray(refs)

    if order == 'xy':
        rfarr = rfarr[:, ::-1]

    offarr = rfarr - rfarr[0]
    lower_corners = -offarr  # coordinate of the lower corner

    # Coordinates with respect to 0
    baseref = lower_corners.min(axis=0)

    # With respect to baseref
    lower_corners = lower_corners - baseref
    upper_corners = sharr - offarr - baseref
    finalshape = tuple(upper_corners.max(axis=0))

    subshapes = []
    for lc, uc in zip(lower_corners, upper_corners):
        subshape = []
        for start, end in zip(lc, uc):
            subshape.append(slice(start, end))
        subshapes.append(tuple(subshape))

    # Coordinate of the refs pixels in final image
    # It's equal for all pixels
    ref_final_0 = lower_corners[0] + rfarr[0]

    if order == 'xy':
        ref_final_0 = ref_final_0[::-1]

    return finalshape, subshapes, tuple(ref_final_0)


def resize_array(data, finalshape, region, window=None,
                 scale=1, fill=0.0, conserve=True, dtype=None):

    if window is not None:
        data = data[window]

    if scale == 1:
        finaldata = data
    else:
        finaldata = rebin_scale(data, scale)

    if dtype is None:
        dtype = data.dtype

    newdata = numpy.empty(finalshape, dtype=dtype)
    newdata.fill(fill)
    newdata[region] = finaldata
    # Conserve the total sum of the original data
    if conserve:
        newdata[region] = newdata[region] / scale**2
    return newdata


def resize_arrays(arrays, shape, offsetsp, finalshape, window=None, scale=1, conserve=True, fill=0.0):
    rarrays = []
    regions = []
    for array, rel_offset in zip(arrays, offsetsp):
        region, _ = subarray_match(finalshape, rel_offset, shape)
        newdata = resize_array(array, finalshape, region, window=window,
                               fill=fill, scale=scale, conserve=conserve)
        rarrays.append(newdata)
        regions.append(region)
    return rarrays, regions


def resize_arrays_alt(arrays, regions, finalshape, window=None, scale=1, conserve=True, fill=0.0):
    rarrays = []
    for array, region in zip(arrays, regions):
        newdata = resize_array(array, finalshape, region, window=window,
                               fill=fill, scale=scale, conserve=conserve)
        rarrays.append(newdata)
    return rarrays


def rebin_scale(a, scale=1):
    """Scale an array to a new shape."""

    newshape = tuple((side * scale) for side in a.shape)

    return rebin(a, newshape)


def rebin(a, newshape):
    """Rebin an array to a new shape."""

    slices = [slice(0, old, float(old)/new)
              for old, new in zip(a.shape, newshape)]
    coordinates = numpy.mgrid[slices]
    # choose the biggest smaller integer index
    indices = coordinates.astype('i')
    return a[tuple(indices)]


def fixpix(data, mask, kind='linear'):
    """Interpolate 2D array data in rows"""
    if data.shape != mask.shape:
        raise ValueError

    if not numpy.any(mask):
        return data

    x = numpy.arange(0, data.shape[0])
    for row, mrow in zip(data, mask):
        if numpy.any(mrow):  # Interpolate if there's some pixel missing
            valid = (mrow == numpy.False_)
            invalid = (mrow == numpy.True_)
            itp = interp1d(x[valid], row[valid], kind=kind, copy=False)
            row[invalid] = itp(x[invalid]).astype(row.dtype)
    return data


def fixpix2(data, mask, iterations=3, out=None):
    """Substitute pixels in mask by a bilinear least square fitting.
    """
    out = out if out is not None else data.copy()

    # A binary mask, regions are ones
    binry = mask != 0

    # Label regions in the binary mask
    lblarr, labl = ndimage.label(binry)

    # Structure for dilation is 8-way
    stct = ndimage.generate_binary_structure(2, 2)
    # Pixels in the background
    back = lblarr == 0
    # For each object
    for idx in range(1, labl + 1):
        # Pixels of the object
        segm = lblarr == idx
        # Pixels of the object or the background
        # dilation will only touch these pixels
        dilmask = numpy.logical_or(back, segm)
        # Dilation 3 times
        more = ndimage.binary_dilation(segm, stct,
                                       iterations=iterations,
                                       mask=dilmask)
        # Border pixels
        # Pixels in the border around the object are
        # more and (not segm)
        border = numpy.logical_and(more, numpy.logical_not(segm))
        # Pixels in the border
        xi, yi = border.nonzero()
        # Bilinear leastsq calculator
        calc = FitOne(xi, yi, out[xi, yi])
        # Pixels in the region
        xi, yi = segm.nonzero()
        # Value is obtained from the fit
        out[segm] = calc(xi, yi)

    return out


def correct_dark(data, dark, dtype='float32'):
    result = data - dark
    result = result.astype(dtype)
    return result


def correct_flatfield(data, flat, dtype='float32'):
    result = data / flat
    result = result.astype(dtype)
    return result


def correct_sky(data, sky, dtype='float32'):
    result = data - sky
    result = result.astype(dtype)
    return result


def correct_nonlinearity(data, polynomial, dtype='float32'):
    result = numpy.polyval(polynomial, data)
    result = result.astype(dtype)
    return result


def compute_sky_advanced(data, omasks):
    d = data[0]
    m = omasks[0]
    median_sky = numpy.median(d[m == 0])
    result = numpy.zeros(data[0].shape)
    result += median_sky
    return result


def compute_median_background(img, omask, region):
    d = img[region]
    m = omask[region]
    median_sky = numpy.median(d[m == 0])
    return median_sky


def numberarray(x, shape):
    """Return x if it is an array or create an array and fill it with x."""
    try:
        iter(x)
    except TypeError:
        return numpy.ones(shape) * x
    else:
        return x



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from astropy.io import fits
import numpy as np

from .define_3d_wcs import wcs_to_header_using_cd_keywords

def generate_image3d_method0_ifu(
        wcs3d,
        header_keys,
        simulated_x_ifu_all,
        simulated_y_ifu_all,
        simulated_wave_all,
        bins_x_ifu,
        bins_y_ifu,
        bins_wave,
        prefix_intermediate_fits
):
    """Compute 3D image3 IFU, method 0

    The image is calculated by performing a three-dimensional histogram
    over the X, Y coordinates in the IFU and the wavelength of the
    simulated photons.

    The result is saved as a FITS file.

    Parameters
    ----------
    wcs3d : `~astropy.wcs.wcs.WCS`
        WCS of the data cube.
    header_keys : `~astropy.io.fits.header.Header`
        FITS header with additional keywords to be merged together with
        the WCS information.
    simulated_x_ifu_all : `~astropy.units.Quantity`
        Simulated X coordinates of the photons in the IFU.
    simulated_y_ifu_all : `~astropy.units.Quantity`
        Simulated Y coordinates of the photons in the IFU.
    simulated_wave_all : `~astropy.units.Quantity`
        Simulated wavelengths of the photons in the IFU.
    bins_x_ifu : `~numpy.ndarray`
        Bin edges in the naxis1_ifu direction
        (along the slice).
    bins_y_ifu : `~numpy.ndarray`
        Bin edges in the naxis2_ifu direction
        (perpendicular to the slice).
    bins_wave : `~numpy.ndarray`
        Bin edges in the wavelength direction.
    prefix_intermediate_fits : str
        Prefix for output intermediate FITS files. If the length of
        this string is 0, no output is generated.

    """

    # generate image
    image3d_method0_ifu, edges = np.histogramdd(
        sample=(simulated_wave_all.value, simulated_y_ifu_all.value, simulated_x_ifu_all.value),
        bins=(bins_wave.value, bins_y_ifu.value, bins_x_ifu.value)
    )

    # save FITS file
    if len(prefix_intermediate_fits) > 0:
        hdu = fits.PrimaryHDU(image3d_method0_ifu.astype(np.uint16))
        pos0 = len(hdu.header) - 1
        hdu.header.extend(wcs_to_header_using_cd_keywords(wcs3d), update=True)
        hdu.header.update(header_keys)
        hdu.header.insert(
            pos0, ('COMMENT', "FITS (Flexible Image Transport System) format is defined in 'Astronomy"))
        hdu.header.insert(
            pos0 + 1, ('COMMENT', "and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H"))
        hdul = fits.HDUList([hdu])
        outfile = f'{prefix_intermediate_fits}_ifu_3D_method0.fits'
        print(f'Saving file: {outfile}')
        hdul.writeto(f'{outfile}', overwrite='yes')



#
# Copyright 2015-2021 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

#cython: language_level=3

from libc.math cimport floor
import numpy
cimport numpy

cdef int wc_to_pix2(double x):
    return <int>(floor(x + 0.5))

ctypedef fused FType:
    double
    float
    int
    long

ctypedef fused IType:
    int
    long


cdef inline int int_max(int a, int b): return a if a >= b else b
cdef inline int int_min(int a, int b): return a if a <= b else b

def extract_simple_intl(FType[:,:] data, IType[:] xx,
                        double[:] bb1, double[:] bb2,
                        double[:] out):

    cdef size_t size = xx.shape[0]
    cdef size_t i
    cdef int pa, pb
    cdef IType x
    cdef double a,b,w,acc
    
    for i in range(size):
        x = xx[i]
        a = bb1[i]
        b = bb2[i]
        pb = int_min(wc_to_pix2(b), data.shape[0])
        pa = int_max(0, wc_to_pix2(a))    
        if pa == pb:
            if pa >= 0 and pa < data.shape[0]:
                w = b - a
                out[x] = data[pa, x] * w
        else:
            acc = 0
            if pa >= 0 and pa < data.shape[0]:
                w = pa + 0.5 - a
                acc += data[pa, x] * w
            
            if pb < data.shape[1] and pb >= 0:
                w = b - (pb -0.5)
                acc += data[pb, x] * w
            for c in range(pa + 1, pb):
                acc += data[c,x]
            out[x] = acc
    return out




#
# Copyright 2008-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""User command line interface of Numina."""


import logging

from .baserun import run_reduce
from .helpers import create_datamanager, load_observations

_logger = logging.getLogger(__name__)


def mode_run_common(args, extra_args, config, mode):
    # FIXME: implement 'recipe' run mode
    if mode == 'rec':
        print('Mode not implemented yet')
        return 1
    elif mode == 'obs':
        return mode_run_common_obs(args, extra_args, config)
    else:
        raise ValueError(f'Not valid run mode {mode}')


def mode_run_common_obs(args, extra_args, config):
    """Observing mode processing mode of numina."""

    # Loading observation result if exists
    sessions, loaded_obs = load_observations(args.obsresult, args.session)

    # Override like this
    if args.basedir:
        config['tool.run']['basedir'] = args.basedir
    if args.datadir:
        config['tool.run']['datadir'] = args.datadir
    if args.copy_files:
        config['tool.run']['copy_files'] = str(args.copy_files)
    if args.validate:
        config['tool.run']['validate'] = str(args.validate)
    if hasattr(args, "strict_reqs"):
        config['tool.run']['strict_reqs'] = str(args.strict_reqs)

    datamanager = create_datamanager(config, args.reqs, extra_args.extra_control)
    datamanager.backend.add_obs(loaded_obs)

    # Start processing
    jobs = []
    for session in sessions:
        for job in session:
            if job['enabled'] or job['id'] in args.enable:
                jobs.append(job)

    copy_files = config['tool.run'].getboolean('copy_files')
    validate = config['tool.run'].getboolean('validate')
    strict_inputs = config['tool.run'].getboolean('strict_reqs')
    try:
        for job in jobs:
            run_reduce(
                datamanager, job['id'], copy_files=copy_files,
                validate_inputs=validate, validate_results=validate,
                strict_inputs=strict_inputs
            )
    finally:
        if args.dump_control:
            _logger.debug('dump control status')
            with open('control_dump.yaml', 'w') as fp:
                datamanager.backend.dump(fp)



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""User command line interface of Numina, show-instruments functionallity."""

import numina.drps
import numina.util.fqn as objimport
import numina.instrument.assembly as asbl


def register(subparsers, config):
    parser_show_ins = subparsers.add_parser(
        'show-instruments',
        help='show registered instruments'
    )

    parser_show_ins.set_defaults(command=show_instruments,
                                 verbose=0, what='om')
    parser_show_ins.add_argument(
        '-o', '--observing-modes',
        action='store_true', dest='om',
        help='list observing modes of each instrument')
    # TODO: this could be the same option for run
    parser_show_ins.add_argument(
        '--profile-path', dest='profilepath',
        default=None,
        help='location of the instrument profiles'
    )
#    parser_show_ins.add_argument('--verbose', '-v', action='count')

    parser_show_ins.add_argument(
        'name', nargs='*', default=None,
        help='filter instruments by name'
    )

    return parser_show_ins


def show_instruments(args, extra_args, config):

    sys_drps = numina.drps.get_system_drps()
    prof_store = asbl.load_panoply_store(sys_drps, args.profilepath)

    if args.name:
        for name in args.name:
            try:
                drp = sys_drps.query_by_name(name)
                print_instrument(drp, prof_store, modes=args.om)
            except KeyError:
                print_no_instrument(name)
    else:
        drps = sys_drps.query_all()
        for drp in drps.values():
            print_instrument(drp, prof_store, modes=args.om)


def print_instrument(instrument, prof_store, modes=True):
    print('Instrument:', instrument.name)
    print(f" version is '{instrument.version}'")
    for key, val in prof_store.items():
        etype = val['type']
        name = val['name']
        if etype == 'instrument' and name == instrument.name:
            desc = val['description']
            uuid = val['uuid']
            msg = f" has profile '{desc}' uuid={uuid}"
            print(msg)

    print(
        f" has datamodel '{objimport.fully_qualified_name(instrument.datamodel)}'")
    for _, pl in instrument.pipelines.items():
        print(f' has pipeline {pl.name!r}, version {pl.version}')
    if modes and instrument.modes:
        print(' has observing modes:')
        for mode in instrument.modes.values():
            print(f"  {mode.name!r} ({mode.key})")


def print_no_instrument(name):
    print('No instrument named:', name)



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Broad spectrum (linear wavelength scale) applying a Gaussian"""

import astropy.constants as const
import astropy.units as u
import numpy as np
from scipy.ndimage import gaussian_filter1d


def broadsp_gaussian_velocity_linearwv(crval1, cdelt1, flux, fwhm_velocity, tsigma=5.0):
    """Apply a Gaussian broadening to a spectrum.

    The method assumes that the spectrum is linearly calibrated
    in wavelength. Since the Gaussian function is parametrized by a
    constant FWHM in velocity units, a varying kernel width (in pixels)
    is computed as we move in the wavelength direction.

    Parameters
    ----------
    crval1 : `~astroppy.units.Quantity`
        Wavelength of the first spectrum pixel.
    cdelt1 : `~astroppy.units.Quantity`
        Constant wavelength increment/pixel.
    flux : array_like
        Array-like object containing the spectrum flux
    fwhm_velocity : `~astroppy.units.Quantity`
        FWHM of the Gaussian kernel in velocity units.
    tsigma : float
        Times sigma to extend the computation of the Gaussian
        broadening.

    Returns
    -------
    broadened_flux : ndarray
        Broadened spectrum.

    """

    # initial checks
    if not isinstance(crval1, u.Quantity):
        raise ValueError(f'{crval1=} is not a Quantity')
    if not crval1.unit.is_equivalent(u.m):
        raise ValueError(f"Unexpected units for 'crval1': {crval1.unit}")

    if not isinstance(cdelt1, u.Quantity):
        raise ValueError(f'{cdelt1=} is not a Quantity')
    if not cdelt1.unit.is_equivalent(u.m / u.pix):
        raise ValueError(f"Unexpected units for 'cdelt1': {cdelt1.unit}")

    if not isinstance(fwhm_velocity, u.Quantity):
        raise ValueError(f'{fwhm_velocity=} is not a Quantity')
    if not fwhm_velocity.unit.is_equivalent(u.m / u.s):
        raise ValueError(f"Unexpected units for 'fwhm_velocity': {fwhm_velocity.unit}")

    # generate wavelength array
    flux = np.asarray(flux)
    naxis1 = len(flux)
    wave = crval1 + np.arange(naxis1) * u.pix * cdelt1

    # conversion factor between FWHM and sigma
    factor_fwhm_to_sigma = 1 / (2 * np.sqrt(2 * np.log(2)))

    # varying sigma for each pixel (wavelength units)
    sigma_wave = fwhm_velocity.to(u.m / u.s) / const.c.to(u.m / u.s) * factor_fwhm_to_sigma * wave

    # varying sigma for each pixel (pixel units)
    sigma_pix = (sigma_wave / cdelt1).value

    # apply varying Gaussian broadening
    broadened_flux = apply_gaussian_broadening_linearwv(
        flux=flux,
        sigma_pix=sigma_pix,
        tsigma=tsigma
    )

    return broadened_flux


def apply_gaussian_broadening_linearwv(flux, sigma_pix, tsigma):
    """Apply a Gaussian broadening to a data array.

    The value of the Gaussian sigma is arbitrary for each pixel.
    In particular, this function is well suited to convolve a spectrum
    with the Gaussian kernel corresponding to a fixed velocity
    dispersion (which leads to a varying sigma in pixel units).

    When 'sigma_pix' is constant for all the pixels, this function
    returns the same result as scipy.ndimage.gaussian_filter1d(...)
    when in the latter the parameter 'mode' is set to 'constant'
    (with constant=0).

    Parameters
    ----------
    flux : ndarray
        Data array corresponding to the spectrum to be broadened.
    sigma_pix : ndarray
        Gaussian kernel width (in float pixel units).
    tsigma : float
        Times sigma to extend the computation of the Gaussian
        broadening.

    Returns
    -------
    broadened_flux : ndarray
        Broadened data array.

    """

    naxis1 = len(flux)
    if len(sigma_pix) != naxis1:
        raise ValueError(f'Incompatible array sizes: {len(flux)=}, {len(sigma_pix)=}')

    # broaden spectrum
    broadened_flux = np.zeros(naxis1)
    for i in range(naxis1):
        # kernel size
        initial_kernel_size_intpix = (2 * sigma_pix[i] * tsigma).astype(int) + 1
        if initial_kernel_size_intpix % 2 == 0:
            kernel_size = initial_kernel_size_intpix + 1
        else:
            kernel_size = initial_kernel_size_intpix
        # location of the kernel center
        kcenter = kernel_size // 2
        # generate array of zeros an insert 1.0 in the central pixel
        delta_filter = np.zeros(kernel_size)
        delta_filter[kcenter] = 1.0
        # generate Gaussian kernel
        gauss_filter = gaussian_filter1d(delta_filter, sigma_pix[i])
        # redistribute the signal from the current pixel to the
        # neighboring pixels, following the kernel calculated in the
        # current pixel
        redistributed_pixel = gauss_filter * flux[i]
        i1 = max(0, i - kcenter)
        i2 = min(naxis1, i + kcenter + 1)
        ii1 = max(0, kcenter - i)
        ii2 = ii1 + (i2 - i1)
        broadened_flux[i1:i2] += redistributed_pixel[ii1:ii2]

    return broadened_flux




import pkgutil

import numina.drps
import numina.drps.drpbase
import numina.core.pipelineload as pload

from ..cli import main


drpdata1 = pkgutil.get_data('numina.drps.tests', 'drptest1.yaml')
drpdata2 = pkgutil.get_data('numina.drps.tests', 'drptest2.yaml')


# def test_show_recipes(capsys, monkeypatch):
def test_show_recipes(capsys, monkeypatch):
    """Test that one instrument is shown"""

    def mockreturn():
        import numina.drps.drpbase
        import numina.core.pipelineload as pload
        drps = {}

        drp1 = pload.drp_load_data('numina', drpdata1)
        drps[drp1.name] = drp1
        return numina.drps.drpbase.DrpGeneric(drps)

    monkeypatch.setattr(numina.drps, "get_system_drps", mockreturn)

    expected = ["Recipe: numina.core.utils.AlwaysFailRecipe",
                " summary: A Recipe that always fails.",
                " instrument: TEST1",
                "  pipeline: default",
                "  obs mode: fail",
                " requirements:",
                "",
                "Recipe: numina.tests.recipes.BiasRecipe",
                " instrument: TEST1",
                "  pipeline: default",
                "  obs mode: bias",
                " requirements:",
                "",
                "Recipe: numina.tests.recipes.DarkRecipe",
                " instrument: TEST1",
                "  pipeline: default",
                "  obs mode: dark",
                " requirements:",
                "",
                "Recipe: numina.tests.recipes.ImageRecipe",
                " instrument: TEST1",
                "  pipeline: default",
                "  obs mode: image",
                " requirements:",
                "  obresult type='ObservationResultType()' [Observation Result]",
                "  master_bias type='MasterBias' [Master Bias]",
                "  master_dark type='MasterDark' [Master Dark]",
                "",
                ""
                ]

    main(['show-recipes'])
    out, err = capsys.readouterr()
    out = out.split("\n")
    out.sort()
    expected.sort()
    print('NN')
    print(out)
    print(expected)
    assert out == expected


def test_show_2_instruments(capsys, monkeypatch):
    """Test that two instruments are shown"""

    def mockreturn():
        drps = {}

        drp1 = pload.drp_load_data('numina', drpdata1)
        drp2 = pload.drp_load_data('numina', drpdata2)
        drps[drp1.name] = drp1
        drps[drp2.name] = drp2
        return numina.drps.drpbase.DrpGeneric(drps)

    monkeypatch.setattr(numina.drps, "get_system_drps", mockreturn)

    expected = [u'Recipe: numina.core.utils.AlwaysFailRecipe',
                u' summary: A Recipe that always fails.',
                u' instrument: TEST1',
                u'  pipeline: default',
                u'  obs mode: fail',
                u' requirements:',
                u'',
                u'Recipe: numina.tests.recipes.BiasRecipe',
                u' instrument: TEST1',
                u'  pipeline: default',
                u'  obs mode: bias',
                u' requirements:',
                u'',
                u'Recipe: numina.tests.recipes.DarkRecipe',
                u' instrument: TEST1',
                u'  pipeline: default',
                u'  obs mode: dark',
                u' requirements:',
                u'',
                u"Recipe: numina.tests.recipes.ImageRecipe",
                u" instrument: TEST1",
                u"  pipeline: default",
                u"  obs mode: image",
                u" requirements:",
                u"  obresult type='ObservationResultType()' [Observation Result]",
                u"  master_bias type='MasterBias' [Master Bias]",
                u"  master_dark type='MasterDark' [Master Dark]",
                u"",
                u'Recipe: numina.tests.recipes.DarkRecipe',
                u' instrument: TEST2',
                u'  pipeline: default',
                u'  obs mode: dark',
                u' requirements:',
                u'',
                u'Recipe: numina.core.utils.AlwaysSuccessRecipe',
                u' summary: A Recipe that always successes.',
                u' instrument: TEST2',
                u'  pipeline: default',
                u'  obs mode: success',
                u' requirements:',
                u'',
                u''
                ]

    main(['show-recipes'])

    out, err = capsys.readouterr()
    out = out.split("\n")
    out.sort()
    expected.sort()
    assert out == expected


def test_show_1_recipe(capsys, monkeypatch):
    """Test that one recipe is shown"""

    def mockreturn():
        drps = {}

        drp1 = pload.drp_load_data('numina', drpdata1)
        drp2 = pload.drp_load_data('numina', drpdata2)
        drps[drp1.name] = drp1
        drps[drp2.name] = drp2
        return numina.drps.drpbase.DrpGeneric(drps)

    monkeypatch.setattr(numina.drps, "get_system_drps", mockreturn)

    expected = [u'Recipe: numina.tests.recipes.BiasRecipe',
                u' instrument: TEST1',
                u'  pipeline: default',
                u'  obs mode: bias',
                u' requirements:',
                u'',
                u''
                ]

    main(['show-recipes', 'numina.tests.recipes.BiasRecipe'])
    out, err = capsys.readouterr()
    out = out.split("\n")
    out.sort()
    expected.sort()
    assert out == expected


def test_show_2_recipes(capsys, monkeypatch):
    """Test that two recipes are shown"""

    def mockreturn():
        drps = {}

        drp1 = pload.drp_load_data('numina', drpdata1)
        # drp2 = pload.drp_load_data('numina', drpdata2)
        drps[drp1.name] = drp1
        # drps[drp2.name] = drp2
        return numina.drps.drpbase.DrpGeneric(drps)

    monkeypatch.setattr(numina.drps, "get_system_drps", mockreturn)

    expected = [u'Recipe: numina.tests.recipes.BiasRecipe',
                u' instrument: TEST1',
                u'  pipeline: default',
                u'  obs mode: bias',
                u' requirements:',
                u'',
                u'Recipe: numina.tests.recipes.DarkRecipe',
                u' instrument: TEST1',
                u'  pipeline: default',
                u'  obs mode: dark',
                u' requirements:',
                u'',
                u''
                ]

    main(
        ['show-recipes',
         'numina.tests.recipes.BiasRecipe',
         'numina.tests.recipes.DarkRecipe'
         ]
    )
    out, err = capsys.readouterr()
    out = out.split("\n")
    out.sort()
    expected.sort()
    assert out == expected


def test_show_2_ins_recipes(capsys, monkeypatch):
    """Test that two recipes are shown, selecting instrument"""

    def mockreturn():
        drps = {}

        drp1 = pload.drp_load_data('numina', drpdata1)
        drp2 = pload.drp_load_data('numina', drpdata2)
        drps[drp1.name] = drp1
        drps[drp2.name] = drp2
        return numina.drps.drpbase.DrpGeneric(drps)

    monkeypatch.setattr(numina.drps, "get_system_drps", mockreturn)

    expected = [u'Recipe: numina.tests.recipes.BiasRecipe',
                u' instrument: TEST1',
                u'  pipeline: default',
                u'  obs mode: bias',
                u' requirements:',
                u'',
                u'Recipe: numina.tests.recipes.DarkRecipe',
                u' instrument: TEST1',
                u'  pipeline: default',
                u'  obs mode: dark',
                u' requirements:',
                u'',
                u''
                ]

    main(
        ['show-recipes', '-i', 'TEST1',
         'numina.tests.recipes.BiasRecipe',
         'numina.tests.recipes.DarkRecipe'
         ]
    )
    out, err = capsys.readouterr()
    out = out.split("\n")
    out.sort()
    expected.sort()
    assert out == expected


def test_show_mode_name(capsys, monkeypatch):
    """Test that one recipe by mode name is shown"""

    def mockreturn():
        drps = {}

        drp1 = pload.drp_load_data('numina', drpdata1)
        # drp2 = pload.drp_load_data('numina', drpdata2)
        drps[drp1.name] = drp1
        # drps[drp2.name] = drp2
        return numina.drps.drpbase.DrpGeneric(drps)

    monkeypatch.setattr(numina.drps, "get_system_drps", mockreturn)

    expected = [u'Recipe: numina.tests.recipes.BiasRecipe',
                u' instrument: TEST1',
                u'  pipeline: default',
                u'  obs mode: bias',
                u' requirements:',
                u'',
                u''
                ]

    main(['show-recipes', '--mode', 'bias'])
    out, err = capsys.readouterr()
    out = out.split("\n")
    out.sort()
    expected.sort()
    assert out == expected


def test_show_recipes_no_instruments(capsys, monkeypatch):
    """Test that no instruments are shown"""

    def mockreturn():
        return numina.drps.drpbase.DrpGeneric()

    monkeypatch.setattr(numina.drps, "get_system_drps", mockreturn)

    expected = ""

    main(['show-recipes'])

    out, err = capsys.readouterr()

    assert out == expected


def test_show_recipes_2_instruments_select_no(capsys, monkeypatch):
    """Test that two instruments are shown"""

    def mockreturn():
        drps = {}
        drp1 = pload.drp_load_data('numina', drpdata1)
        drp2 = pload.drp_load_data('numina', drpdata2)
        drps[drp1.name] = drp1
        drps[drp2.name] = drp2
        return numina.drps.drpbase.DrpGeneric(drps)

    monkeypatch.setattr(numina.drps, "get_system_drps", mockreturn)

    expected = ["No instrument named: TEST3",
                ""
                ]

    main(['show-recipes', '-i', 'TEST3'])

    out, err = capsys.readouterr()
    out = out.split("\n")
    out.sort()
    expected.sort()
    assert out == expected


def test_show_recipes_no_instruments_select_no(capsys, monkeypatch):
    """Test that two instruments are shown"""

    def mockreturn():
        drps = {}
        return numina.drps.drpbase.DrpGeneric(drps)

    monkeypatch.setattr(numina.drps, "get_system_drps", mockreturn)

    expected = ["No instrument named: TEST3",
                ""
                ]

    main(['show-recipes', '-i', 'TEST3'])

    out, err = capsys.readouterr()
    out = out.split("\n")
    out.sort()
    expected.sort()
    assert out == expected



#
# Copyright 2017-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Test convert strings to functions in data load """

import numpy
import numpy.polynomial.polynomial as nppol

from numina.util.convertfunc import json_serial_function


def test_convert_poly1():
    coef = [1.0, 0.4, -6.1]
    poly1 = nppol.Polynomial(coef)
    result = json_serial_function(poly1)
    assert result['function'] == 'polynomial1d'
    assert numpy.allclose(result['params'], coef)




import pytest
import pkgutil

from numina.drps.drpsystem import DrpSystem


def test_drpsys_one_instrument(drpmocker):
    """Test that only one DRP is returned."""

    drpdata1 = pkgutil.get_data('numina.drps.tests', 'drptest1.yaml')

    drpmocker.add_drp('TEST1', drpdata1)

    drpsys = DrpSystem()
    drpsys.load()

    ldrp = drpsys.query_by_name('TEST1')

    assert ldrp is not None
    assert ldrp.name == 'TEST1'

    with pytest.raises(KeyError):
        drpsys.query_by_name('OTHER')

    alldrps = drpsys.query_all()
    assert len(alldrps) == 1
    assert 'TEST1' in alldrps
    # FIXME: We should check that both are equal, not just the name
    assert alldrps['TEST1'].name == ldrp.name


def test_drpsys_2_instruments(drpmocker):
    """Test that two DRPs are returned"""

    drpdata1 = pkgutil.get_data('numina.drps.tests', 'drptest1.yaml')
    drpdata2 = pkgutil.get_data('numina.drps.tests', 'drptest2.yaml')
    drpmocker.add_drp('TEST1', drpdata1)
    drpmocker.add_drp('TEST2', drpdata2)

    drpsys = DrpSystem()
    drpsys.load()

    ldrp1 = drpsys.query_by_name('TEST1')

    assert ldrp1 is not None
    assert ldrp1.name == 'TEST1'

    ldrp2 = drpsys.query_by_name('TEST2')

    assert ldrp2 is not None
    assert ldrp2.name == 'TEST2'

    with pytest.raises(KeyError):
        drpsys.query_by_name('OTHER')

    alldrps = drpsys.query_all()
    assert len(alldrps) == 2
    assert 'TEST1' in alldrps
    # FIXME: We should check that both are equal, not just the name
    assert alldrps['TEST1'].name == ldrp1.name

    assert 'TEST2' in alldrps
    # FIXME: We should check that both are equal, not just the name
    assert alldrps['TEST2'].name == ldrp2.name


def test_drpsys_name_2_instruments(drpmocker):
    """Test that two DRPs are returned"""

    drpdata1 = pkgutil.get_data('numina.drps.tests', 'drptest1.yaml')
    drpdata2 = pkgutil.get_data('numina.drps.tests', 'drptest2.yaml')
    drpmocker.add_drp('TEST1', drpdata1)
    drpmocker.add_drp('TEST2', drpdata2)

    ldrp1 = DrpSystem.load_drp('TEST1')

    assert ldrp1 is not None
    assert ldrp1.name == 'TEST1'

    ldrp2 = DrpSystem.load_drp('TEST2')

    assert ldrp2 is not None
    assert ldrp2.name == 'TEST2'

    with pytest.raises(KeyError):
        DrpSystem.load_drp('OTHER')


def test_drpsys_iload_2_instruments(drpmocker):
    """Test that two DRPs are returned"""

    drpdata1 = pkgutil.get_data('numina.drps.tests', 'drptest1.yaml')
    drpdata2 = pkgutil.get_data('numina.drps.tests', 'drptest2.yaml')
    drpmocker.add_drp('TEST1', drpdata1)
    drpmocker.add_drp('TEST2', drpdata2)

    gendrp = DrpSystem.iload()

    for ldrp in gendrp:
        assert ldrp is not None


@pytest.mark.usefixtures("drpmocker")
def test_drpsys_no_instrument():

    drpsys = DrpSystem()
    drpsys.load()

    with pytest.raises(KeyError):
        drpsys.query_by_name('TEST1')

    with pytest.raises(KeyError):
        drpsys.query_by_name('TEST2')

    with pytest.raises(KeyError):
        drpsys.query_by_name('OTHER')

    alldrps = drpsys.query_all()
    assert len(alldrps) == 0


def test_drpsys_bad_file(capsys, drpmocker):
    """Test that a bad file doesn't break the load"""

    drpdata3 = pkgutil.get_data('numina.drps.tests', 'drptest3.yaml')

    drpmocker.add_drp('TEST3', drpdata3)
    drpsys = DrpSystem()
    drpsys.load()
    expected_msg = [
        '',
        "Error is:  Missing key 'modes' inside 'root' node",
        "Problem loading EntryPoint(name='TEST3', value='TEST3.loader', group='numina.pipeline.1')"
    ]

    out, err = capsys.readouterr()
    err = err.split("\n")
    err.sort()

    assert err == expected_msg



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from rich.console import Console


# Define NuminaConsole class for enriched output, inheriting from
# the Console class in the rich module
class NuminaConsole(Console):
    """Numina console for user interaction."""

    def __init__(self, **kwargs):
        """Initialize the Numina console with custom settings."""
        if isinstance(kwargs, dict):
            if 'force_terminal' not in kwargs:
                kwargs['force_terminal'] = True
        super().__init__(**kwargs)



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from astropy.coordinates import SkyCoord
import astropy.units as u
from astropy.wcs.utils import proj_plane_pixel_scales
import matplotlib.pyplot as plt
import numpy as np

from numina.tools.ctext import ctext
from numina.tools.compute_adr_wavelength import compute_adr_wavelength

from .raise_valueerror import raise_ValueError
from .simulate_image2d_from_fitsfile import simulate_image2d_from_fitsfile


def generate_geometry_for_scene_block(
        scene_fname, scene_block, nphotons,
        apply_seeing, seeing_fwhm_arcsec, seeing_psf,
        airmass, parallactic_angle,
        reference_wave_vacuum_differential_refraction, simulated_wave,
        instrument_pa,
        wcs3d,
        min_x_ifu, max_x_ifu, min_y_ifu, max_y_ifu,
        rng,
        verbose, plots
):
    """Distribute photons in the IFU focal plane for the scene block.

    Parameters
    ----------
    scene_fname : str
        YAML scene file name.
    scene_block : dict
        Dictonary storing a scene block.
    nphotons : int
        Number of photons to be generated in the scene block.
    apply_seeing : bool
        If True, apply seeing to simulated photons.
    seeing_fwhm_arcsec : `~astropy.units.Quantity`
        Seeing FWHM.
    seeing_psf : str
        Seeing PSF.
    airmass : float
        Airmass.
    parallactic_angle : `~astropy.units.Quantity`
        Parallactic angle. This number must be within the range
        [-90,+90] deg.
    reference_wave_vacuum_differential_refraction : `~astropy.units.Quantity`
        Reference wavelength to compute the differential refraction
        correction. This wavelength corresponds to a correction of
        zero.
    simulated_wave : `~astropy.units.Quantity`
        Array containing `nphotons` simulated photons with the
        spectrum requested in the scene block. These values are
        required to compute the differential refraction correction.
    instrument_pa : `~astropy.units.Quantity`
        Instrument position angle.
    wcs3d : `~astropy.wcs.wcs.WCS`
        WCS of the data cube.
    min_x_ifu : `~astropy.units.Quantity`
        Minimum pixel X coordinate defining the IFU focal plane.
    max_x_ifu : `~astropy.units.Quantity`
        Maximum pixel X coordinate defining the IFU focal plane.
    min_y_ifu : `~astropy.units.Quantity`
        Minimum pixel Y coordinate defining the IFU focal plane.
    max_y_ifu : `~astropy.units.Quantity`
        Maximum pixel Y coordinate defining the IFU focal plane.
    rng : `~numpy.random._generator.Generator`
        Random number generator.
    verbose : bool
        If True, display additional information.
    plots : bool
        If True, plot intermediate results.

    Returns
    -------
    simulated_x_ifu_corrected : `~astropy.units.Quantity`
        Array containing the X coordinate of the 'nphotons' photons
        in the focal plane of the IFU.
    simulated_y_ifu_corrected : `~astropy.units.Quantity`
        Array containing the X coordinate of the 'nphotons' photons
        in the focal plane of the IFU.

    """

    if len(simulated_wave) != nphotons:
        raise ValueError(f'Unexpected {len(simulated_wave)=} != {nphotons=}')

    factor_fwhm_to_sigma = 1 / (2 * np.sqrt(2 * np.log(2)))

    # plate scale
    plate_scale_x, plate_scale_y, plate_scale_wave = proj_plane_pixel_scales(wcs3d)
    plate_scale_x *= u.deg / u.pix
    plate_scale_y *= u.deg / u.pix

    if verbose:
        print(f'{wcs3d.wcs.cd=}')
        print(f'{plate_scale_x=}')
        print(f'{plate_scale_y=}')

    # define geometry type for scene block
    geometry_type = scene_block['geometry']['type']

    # simulate photons following the selected geometry
    if geometry_type == 'flatfield':
        simulated_x_ifu = rng.uniform(low=min_x_ifu.value, high=max_x_ifu.value, size=nphotons)
        simulated_y_ifu = rng.uniform(low=min_y_ifu.value, high=max_y_ifu.value, size=nphotons)
    elif geometry_type in ['gaussian', 'point-like', 'from-FITS-image']:
        if 'ra_deg' in scene_block['geometry']:
            ra_deg = scene_block['geometry']['ra_deg']
        else:
            if verbose:
                print(ctext('Assuming ra_deg: 0', faint=True))
            ra_deg = 0.0
        ra_deg *= u.deg
        if 'dec_deg' in scene_block['geometry']:
            dec_deg = scene_block['geometry']['dec_deg']
        else:
            if verbose:
                print(ctext('Assuming dec_deg: 0', faint=True))
            dec_deg = 0.0
        dec_deg *= u.deg
        if 'delta_ra_arcsec' in scene_block['geometry']:
            delta_ra_arcsec = scene_block['geometry']['delta_ra_arcsec']
        else:
            if verbose:
                print(ctext('Assuming delta_ra_deg: 0', faint=True))
            delta_ra_arcsec = 0.0
        delta_ra_arcsec *= u.arcsec
        if 'delta_dec_arcsec' in scene_block['geometry']:
            delta_dec_arcsec = scene_block['geometry']['delta_dec_arcsec']
        else:
            if verbose:
                print(ctext('Assuming delta_dec_deg: 0', faint=True))
            delta_dec_arcsec = 0.0
        delta_dec_arcsec *= u.arcsec
        x_center, y_center = wcs3d.celestial.world_to_pixel(
            SkyCoord(ra=ra_deg + delta_ra_arcsec.to(u.deg), dec=dec_deg + delta_dec_arcsec.to(u.deg))
        )
        # the previous pixel coordinates are assumed to be 0 at the center
        # of the first pixel in each dimension
        x_center += 1
        y_center += 1
        if geometry_type == 'point-like':
            simulated_x_ifu = np.repeat(x_center, nphotons)
            simulated_y_ifu = np.repeat(y_center, nphotons)
        elif geometry_type == 'gaussian':
            mandatory_keys = ['fwhm_ra_arcsec']
            for key in mandatory_keys:
                if key not in scene_block['geometry']:
                    raise_ValueError(f"Expected key '{key}' not found!")
            fwhm_ra_arcsec = scene_block['geometry']['fwhm_ra_arcsec'] * u.arcsec
            if 'fwhm_dec_arcsec' in scene_block['geometry']:
                fwhm_dec_arcsec = scene_block['geometry']['fwhm_dec_arcsec'] * u.arcsec
            else:
                fwhm_dec_arcsec = fwhm_ra_arcsec
                if verbose:
                    print(ctext(f'Assuming {fwhm_dec_arcsec=}', faint=True))
            if 'position_angle_deg' in scene_block['geometry']:
                position_angle_deg = scene_block['geometry']['position_angle_deg'] * u.deg
            else:
                position_angle_deg = 0.0 * u.deg
                if verbose:
                    print(ctext(f'Assuming {position_angle_deg=}', faint=True))
            # covariance matrix for the multivariate normal
            std_x = fwhm_ra_arcsec * factor_fwhm_to_sigma / plate_scale_x.to(u.arcsec / u.pix)
            std_y = fwhm_dec_arcsec * factor_fwhm_to_sigma / plate_scale_y.to(u.arcsec / u.pix)
            rotation_matrix = np.array(  # note the sign to rotate N -> E -> S -> W
                [
                    [np.cos(position_angle_deg), np.sin(position_angle_deg)],
                    [-np.sin(position_angle_deg), np.cos(position_angle_deg)]
                ]
            )
            covariance = np.diag([std_x.value ** 2, std_y.value ** 2])
            rotated_covariance = np.dot(rotation_matrix.T, np.dot(covariance, rotation_matrix))
            # simulate X, Y values
            simulated_xy_ifu = rng.multivariate_normal(
                mean=[x_center, y_center],
                cov=rotated_covariance,
                size=nphotons
            )
            # compensate for instrument position angle
            simulated_x_ifu = \
                simulated_xy_ifu[:, 0] * np.cos(instrument_pa).value + \
                simulated_xy_ifu[:, 1] * np.sin(instrument_pa).value
            simulated_y_ifu = \
                -simulated_xy_ifu[:, 0] * np.sin(instrument_pa).value + \
                simulated_xy_ifu[:, 1] * np.cos(instrument_pa).value
        elif geometry_type == 'from-FITS-image':
            mandatory_keys = ['filename', 'diagonal_fov_arcsec', 'background_to_subtract']
            for key in mandatory_keys:
                if key not in scene_block['geometry']:
                    raise_ValueError(f"Expected key '{key}' not found!")
            # read reference FITS file
            infile = scene_block['geometry']['filename']
            diagonal_fov_arcsec = scene_block['geometry']['diagonal_fov_arcsec'] * u.arcsec
            background_to_subtract = scene_block['geometry']['background_to_subtract']
            # generate simulated locations in the IFU
            simulated_x_ifu_0, simulated_y_ifu_0 = simulate_image2d_from_fitsfile(
                infile=infile,
                diagonal_fov_arcsec=diagonal_fov_arcsec,
                plate_scale_x=plate_scale_x,
                plate_scale_y=plate_scale_y,
                nphotons=nphotons,
                rng=rng,
                background_to_subtract=background_to_subtract,
                plots=plots,
                verbose=verbose
            )
            # compensate for instrument rotation angle
            simulated_x_ifu = \
                simulated_x_ifu_0 * np.cos(instrument_pa).value + \
                simulated_y_ifu_0 * np.sin(instrument_pa).value
            simulated_y_ifu = \
                -simulated_x_ifu_0 * np.sin(instrument_pa).value + \
                simulated_y_ifu_0 * np.cos(instrument_pa).value
            # shift image center
            simulated_x_ifu += x_center
            simulated_y_ifu += y_center
        else:
            simulated_x_ifu = None  # avoid PyCharm warning (not aware of raise ValueError)
            simulated_y_ifu = None  # avoid PyCharm warning (not aware of raise ValueError)
            raise_ValueError(f'Unexpected {geometry_type=}')
    else:
        simulated_x_ifu = None  # avoid PyCharm warning (not aware of raise ValueError)
        simulated_y_ifu = None  # avoid PyCharm warning (not aware of raise ValueError)
        raise_ValueError(f'Unexpected {geometry_type=} in file {scene_fname}')

    # apply seeing
    if apply_seeing:
        if seeing_psf == "gaussian":
            if verbose:
                print(f'Applying Gaussian PSF with {seeing_fwhm_arcsec=}')
            std_x = seeing_fwhm_arcsec * factor_fwhm_to_sigma / plate_scale_x.to(u.arcsec / u.pix)
            simulated_x_ifu += rng.normal(loc=0, scale=abs(std_x.value), size=nphotons)
            std_y = seeing_fwhm_arcsec * factor_fwhm_to_sigma / plate_scale_y.to(u.arcsec / u.pix)
            simulated_y_ifu += rng.normal(loc=0, scale=abs(std_y.value), size=nphotons)
        else:
            raise_ValueError(f'Unexpected {seeing_psf=}')

    # apply differential refraction (as a function of wavelength)
    if (geometry_type != 'flatfield') and (airmass > 1.0):
        if verbose:
            print('Applying differential refraction correction as a function of wavelength')
        differential_refraction = compute_adr_wavelength(
            airmass=airmass,
            reference_wave_vacuum=reference_wave_vacuum_differential_refraction,
            wave_vacuum=simulated_wave,
            debug=verbose
        )
        # compute RA and DEC of each simulated photon
        simulated_coor = wcs3d.celestial.pixel_to_world(
            simulated_x_ifu - 1.0,
            simulated_y_ifu - 1.0
        )
        simulated_ra = simulated_coor.ra
        simulated_dec = simulated_coor.dec
        # apply differential refraction correction (first declination
        # and then right ascension; see Eq. (39) and (40), pp. 71-72,
        # in Textbook on Spherical Astronomy, Smart, 1977).
        simulated_dec += differential_refraction.to(u.deg) * np.cos(parallactic_angle)
        simulated_ra += differential_refraction.to(u.deg) * np.sin(parallactic_angle) / np.cos(simulated_dec)
        # recompute X, Y coordinates in the IFU focal plane
        simulated_x_ifu_corrected, simulated_y_ifu_corrected = wcs3d.celestial.world_to_pixel(
            SkyCoord(ra=simulated_ra, dec=simulated_dec)
        )
        simulated_x_ifu_corrected += 1
        simulated_y_ifu_corrected += 1
        if plots:
            iok = np.argwhere(simulated_wave > 0 * u.m)
            # differential refraction (as a function of wavelength)
            # computed at the center of the IFU
            wmin_simulated = np.min(simulated_wave[iok])
            wmax_simulated = np.max(simulated_wave[iok])
            nsample_wavelengths = 20
            sample_wavelengths = np.linspace(wmin_simulated, wmax_simulated, nsample_wavelengths)
            differential_refraction_center_ifu = compute_adr_wavelength(
                airmass=airmass,
                reference_wave_vacuum=reference_wave_vacuum_differential_refraction,
                wave_vacuum=sample_wavelengths
            )
            x_center_ifu, y_center_ifu = wcs3d.celestial.wcs.crpix
            simulated_coor = wcs3d.celestial.pixel_to_world(
                x_center_ifu - 1.0,
                y_center_ifu - 1.0
            )
            ra_center_ifu = simulated_coor.ra
            dec_center_ifu = simulated_coor.dec
            ra_center_ifu = np.repeat(ra_center_ifu, nsample_wavelengths)
            dec_center_ifu = np.repeat(dec_center_ifu, nsample_wavelengths)
            dec_center_ifu += differential_refraction_center_ifu.to(u.deg) * np.cos(parallactic_angle)
            ra_center_ifu += differential_refraction_center_ifu.to(u.deg) * np.sin(parallactic_angle) / np.cos(dec_center_ifu)
            x_center_ifu_corrected, y_center_ifu_corrected = wcs3d.celestial.world_to_pixel(
                SkyCoord(ra=ra_center_ifu, dec=dec_center_ifu)
            )
            x_center_ifu_corrected += 1
            y_center_ifu_corrected += 1
            delta_x_center_ifu = x_center_ifu_corrected - x_center_ifu
            delta_y_center_ifu = y_center_ifu_corrected - y_center_ifu
            # differential refraction (as a function of wavelength)
            # computed for each useful simulated photon
            delta_simulated_x_ifu = simulated_x_ifu_corrected - simulated_x_ifu
            delta_simulated_y_ifu = simulated_y_ifu_corrected - simulated_y_ifu
            fig, axarr = plt.subplots(nrows=2, ncols=1, figsize=(6.4, 6.4))
            for iplot in range(2):
                ax = axarr[iplot]
                if iplot == 0:
                    ax.plot(simulated_wave[iok], delta_simulated_x_ifu[iok], ',', label='simulated photons')
                    ax.plot(sample_wavelengths, delta_x_center_ifu, 'r+', label='IFU center')
                    ax.set_ylabel(r'$\Delta$simulated_x_ifu (pixel)')
                else:
                    ax.plot(simulated_wave[iok], delta_simulated_y_ifu[iok], ',', label='simulated photons')
                    ax.plot(sample_wavelengths, delta_y_center_ifu, 'r+', label='IFU center')
                    ax.set_ylabel(r'$\Delta$simulated_y_ifu (pixel)')
                ax.set_xlabel(f'Wavelength ({simulated_wave.unit})')
                ax.axhline(0, linestyle='--', color='gray')
                ax.axvline(reference_wave_vacuum_differential_refraction.value, linestyle=':', color='C1')
                ax.set_title(f'airmass: {airmass}, parallactic angle: {parallactic_angle}')
                ax.legend()
            plt.tight_layout()
            plt.show()
    else:
        simulated_x_ifu_corrected = simulated_x_ifu
        simulated_y_ifu_corrected = simulated_y_ifu

    # add units
    simulated_x_ifu_corrected *= u.pix
    simulated_y_ifu_corrected *= u.pix

    return simulated_x_ifu_corrected, simulated_y_ifu_corrected





from ..fqn import fully_qualified_name


def test_fqn():

    import numina.core.recipes

    expected_name = "numina.core.recipes.BaseRecipe"
    fqn_name = fully_qualified_name(numina.core.recipes.BaseRecipe)

    assert fqn_name == expected_name



#
# Copyright 2008-2020 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""An implementation of hierarchical dictionary."""

import collections.abc


class TreeDict(collections.abc.MutableMapping):
    """
    Hierarchical dictionary
    """

    def __init__(self, *args, **keywords):
        self._data = dict(*args, **keywords)

    def __getitem__(self, key):
        keys = key.split('.')
        return self._rec_getitem(self._data, keys[0], keys[1:])

    def __iter__(self):
        return iter(self._data)

    def __setitem__(self, key, value):
        keys = key.split('.')
        return self._rec_setitem(self._data, value, keys[0], keys[1:])

    def __delitem__(self, key):
        pass

    def __len__(self):
        return 1

    def _rec_setitem(self, node, value, key, rest):
        if rest:
            if key not in node:
                node[key] = type(self)()
            self._rec_setitem(node[key], value, rest[0], rest[1:])
        else:
            node[key] = value
            return

    def _rec_getitem(self, data, key, rest):
        if rest:
            return self._rec_getitem(data[key], rest[0], rest[1:])
        else:
            return data[key]



#
# Copyright 2014-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Utility routines"""

import math

import numpy


def coor_to_pix_1d(w):
    """Return the pixel where a coordinate is located."""
    return int(math.floor(w + 0.5))


def wcs_to_pix(w):
    """

    Parameters
    ----------
    w

    Returns
    -------

    """
    return [coor_to_pix_1d(w1) for w1 in w[::-1]]


def coor_to_pix(w, order='rc'):
    """

    Parameters
    ----------
    w
    order

    Returns
    -------

    """
    if order == 'xy':
        return coor_to_pix(w[::-1], order='rc')
    return [coor_to_pix_1d(w1) for w1 in w]


def wcs_to_pix_np(w):
    """

    Parameters
    ----------
    w

    Returns
    -------

    """
    wnp = numpy.asarray(w)
    mm = numpy.floor(wnp + 0.5)
    return mm[::-1].astype('int')


def slice_create(center, block, start=0, stop=None):
    """
    Return an slice with a symmetric region around center.

    Parameters
    ----------
    center
    block
    start
    stop

    Returns
    -------

    """

    do = coor_to_pix_1d(center - block)
    up = coor_to_pix_1d(center + block)

    lmax = max(start, do)

    if stop is not None:
        h = min(up + 1, stop)
    else:
        h = up + 1

    return slice(lmax, h, 1)


def image_box(center, shape, box):
    """Create a region of size box, around a center in a image of shape."""
    return tuple(slice_create(c, b, stop=s)
                 for c, s, b in zip(center, shape, box))


def image_box2d(x, y, shape, box):
    """

    Parameters
    ----------
    x
    y
    shape
    box

    Returns
    -------

    """
    return image_box((y, x), shape, box)


def extent(sl):
    """

    Parameters
    ----------
    sl

    Returns
    -------

    """
    result = [sl[1].start-0.5, sl[1].stop-0.5, sl[0].start-0.5, sl[0].stop-0.5]
    return result


def expand_slice(s, a, b, start=0, stop=None):
    """Expand a slice on the start/stop limits"""
    n1 = max(s.start - a, start)
    n2 = s.stop + b
    if stop is not None:
        n2 = min(n2, stop)

    return slice(n1, n2, 1)


def expand_region(tuple_of_s, a, b, start=0, stop=None):
    """Apply expend_slice on a tuple of slices"""
    return tuple(expand_slice(s, a, b, start=start, stop=stop)
                 for s in tuple_of_s)



#
# Copyright 2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""User command line interface of Numina, verify functionality."""

import logging


_logger = logging.getLogger(__name__)


def register(subparsers, config):
    parser_identify = subparsers.add_parser(
        'identify',
        help='identify'
    )
    parser_identify.set_defaults(command=identify)
    parser_identify.add_argument(
        'files', nargs='+',
        help='identify files in CL'
    )

    return parser_identify


def identify(args, extra_args, config):

    # This function loads the recipes
    import numina.drps
    import numina.instrument.assembly as asbl
    import astropy.io.fits as fits

    sys_drps = numina.drps.get_system_drps()
    com_store = asbl.load_panoply_store(sys_drps, None)

    for f in args.files:
        with fits.open(f) as hdulist:
            _logger.debug(f'identify {f}')
            try:
                # Determine the instrument name
                hdr = hdulist[0].header
                instrument = hdr['INSTRUME']
                #
                this_drp = sys_drps.query_by_name(instrument)
                _logger.debug('assembly instrument model')
                key, date_obs, keyname = this_drp.select_profile_image(hdulist)
                config = asbl.assembly_instrument(com_store, key, date_obs, by_key=keyname)
                print('Instrument:', instrument)
                print('Instrument profile', config.uuid)

            except Exception as ex:
                print(ex)
    return 0



import numpy
from ..stats import robust_std
from ..stats import summary


def test_robust_std():
    xdata = numpy.arange(20)

    sigmag = robust_std(xdata)

    assert numpy.allclose([sigmag], [7.0423499999999999])


def test_robust_std_2d():
    xdata = numpy.arange(20).reshape((5, 4))

    sigmag = robust_std(xdata)

    assert numpy.allclose([sigmag], [7.0423499999999999])


def test_summary():
    xdata = numpy.arange(50)

    result = summary(xdata)

    expected_result = {
        'npoints': 50,
        'minimum': 0,
        'percentile25': 12.25,
        'median': 24.5,
        'mean': 24.5,
        'percentile75': 36.75,
        'maximum': 49,
        'std': 14.430869689661812,
        'robust_std': 18.161849999999998,
        'percentile15': 7.7741097000000003,
        'percentile84': 41.225890299999996
    }

    for key, val in result.items():
        assert numpy.allclose(result[key], expected_result[key])



#
# Copyright 2008-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""User command line interface of Numina."""

import os

from .clirundal import mode_run_common


def complete_config(config):
    """Complete config with default values"""
    secname = 'tool.run'
    if not config.has_section(secname):
        config.add_section(secname)

    values = {
        'basedir': os.getcwd(),
    }

    for k, v in values.items():
        if not config.has_option(secname, k):
            config.set(secname, k, v)

    return config


def register(subparsers, config):
    complete_config(config)
    parser_run = subparsers.add_parser(
        'run',
        help='process a observation result'
    )

    parser_run.add_argument(
        '-c', '--task-control', dest='reqs',
        help='configuration file of the processing task', metavar='FILE'
    )
    parser_run.add_argument(
        '-r', '--requirements', dest='reqs',
        help='alias for --task-control', metavar='FILE'
    )
    parser_run.add_argument(
        '-i', '--instrument', dest='insconf',
        default=None,
        help='name of an instrument configuration'
    )
    parser_run.add_argument(
        '--profile-path', dest='profilepath',
        default=None,
        help='location of the instrument profiles'
    )
    parser_run.add_argument(
        '-p', '--pipeline', dest='pipe_name',
        default='default', help='name of a pipeline'
    )
    parser_run.add_argument(
        '--basedir', action="store", dest="basedir",
        help='path to create the following directories'
    )
    parser_run.add_argument(
        '--datadir', action="store", dest="datadir",
        help='path to directory containing pristine data'
    )
    parser_run.add_argument(
        '--resultsdir', action="store", dest="resultsdir",
        help='path to directory to store results'
    )
    parser_run.add_argument(
        '--workdir', action="store", dest="workdir",
        help='path to directory containing intermediate files'
    )
    parser_run.add_argument(
        '--cleanup', action="store_true", dest="cleanup",
        default=False, help='cleanup workdir on exit [disabled]'
    )
    parser_run.add_argument(
        '--link-files', '--not-copy-files', action="store_const", dest="copy_files", const=False,
        help='do not copy observation result and requirement files'
    )
    parser_run.add_argument(
        '-e', '--enable', action='append', default=[],
        metavar='BLOCKID', help='enable blocks by id'
    )
    parser_run.add_argument(
        '--dump-control', action="store_true",
        help='save the modified task control file'
    )
    parser_run.add_argument(
        '--session', action="store_true",
        help='use the obresult file as a session file'
    )
    parser_run.add_argument(
        '--validate', action="store_const", const=True,
        help='validate inputs and results of recipes'
    )
    group_strict = parser_run.add_mutually_exclusive_group()
    group_strict.add_argument(
        '--strict-reqs', action="store_true",
        help='end the program if a requirement name is not matched by the recipe'
    )
    group_strict.add_argument(
        '--loose-reqs', action="store_false", dest='strict_reqs',
        help='warn if a requirement name is not matched by the recipe'
    )
    parser_run.add_argument(
        'obsresult', nargs='+',
        help='file with the observation result'
    )

    parser_run.set_defaults(command=mode_run_obsmode, strict_reqs=True)

    return parser_run


def mode_run_obsmode(args, extra_args, config):
    mode_run_common(args, extra_args, config, mode='obs')



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""User command line interface of Numina."""

import os

from .clirundal import mode_run_common


def register(subparsers, config):
    parser_run_recipe = subparsers.add_parser(
        'run-recipe',
        help='run a recipe'
    )

    parser_run_recipe.set_defaults(command=mode_run_recipe)
    parser_run_recipe.add_argument(
        '-c', '--task-control', dest='reqs',
        help='configuration file of the processing task', metavar='FILE'
    )
    parser_run_recipe.add_argument(
        '-r', '--requirements', dest='reqs',
        help='alias for --task-control', metavar='FILE'
    )
    parser_run_recipe.add_argument(
        '--obs-res', dest='obsresult',
        help='observation result', metavar='FILE'
    )
    parser_run_recipe.add_argument(
        'recipe',
        help='fqn of the recipe'
    )
    parser_run_recipe.add_argument(
        '--basedir', action="store", dest="basedir",
        default=os.getcwd(),
        help='path to create the following directories'
    )
    parser_run_recipe.add_argument(
        '--datadir', action="store", dest="datadir",
        help='path to directory containing pristine data'
    )
    parser_run_recipe.add_argument(
        '--resultsdir', action="store", dest="resultsdir",
        help='path to directory to store results'
    )
    parser_run_recipe.add_argument(
        '--workdir', action="store", dest="workdir",
        help='path to directory containing intermediate files'
    )
    parser_run_recipe.add_argument(
        '--cleanup', action="store_true", dest="cleanup",
        default=False, help='cleanup workdir on exit [disabled]'
    )

    return parser_run_recipe


def mode_run_recipe(args, extra_args, config):
    mode_run_common(args, extra_args, config, mode='rec')



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


import math

import numpy


def fowler_array(fowlerdata, ti=0.0, ts=0.0, gain=1.0, ron=1.0,
                 badpixels=None, dtype='float64',
                 saturation=65631, blank=0, normalize=False):
    """Loop over the first axis applying Fowler processing.

    *fowlerdata* is assumed to be a 3D numpy.ndarray containing the
    result of a nIR observation in Fowler mode (Fowler and Gatley 1991).
    The shape of the array must be of the form 2N_p x M x N, with N_p being
    the number of pairs in Fowler mode.

    The output signal is just the mean value of the differences between the
    last N_p values (S_i) and the first N_p values (R-i).

    .. math::

        S_F = \\frac{1}{N_p}\\sum\\limits_{i=0}^{N_p-1} S_i - R_i


    If the source has a radiance F, then the measured signal is equivalent
    to:

    .. math::

        S_F = F T_I - F T_S (N_p -1) = F T_E

    being T_I the integration time (*ti*), the time since the first
    productive read to the last productive read for a given pixel and T_S the
    time between samples (*ts*). T_E is the time between correlated reads
    :math:`T_E = T_I - T_S (N_p - 1)`.

    The variance of the signnal is the sum of two terms, one for the readout
    noise:

    .. math::

        \\mathrm{var}(S_{F1}) =\\frac{2\\sigma_R^2}{N_p}

    and other for the photon noise:

    .. math::

        \\mathrm{var}(S_{F2}) = F T_E - F T_S \\frac{1}{3}(N_p-\\frac{1}{N_p})
        = F T_I - F T_S (\\frac{4}{3} N_p -1 -  \\frac{1}{3N_p})


    :param fowlerdata: Convertible to a 3D numpy.ndarray with first axis even
    :param ti: Integration time.
    :param ts: Time between samples.
    :param gain: Detector gain.
    :param ron: Detector readout noise in counts.
    :param badpixels: An optional MxN mask of dtype 'uint8'.
    :param dtype: The dtype of the float outputs.
    :param saturation: The saturation level of the detector.
    :param blank: Invalid values in output are substituted by *blank*.
    :returns: A tuple of (signal, variance of the signal, numper of pixels used
        and badpixel mask.
    :raises: ValueError

    """
    import numina.array._nirproc as _nirproc

    if gain <= 0:
        raise ValueError("invalid parameter, gain <= 0.0")

    if ron <= 0:
        raise ValueError("invalid parameter, ron < 0.0")

    if ti < 0:
        raise ValueError("invalid parameter, ti < 0.0")

    if ts < 0:
        raise ValueError("invalid parameter, ts < 0.0")

    if saturation <= 0:
        raise ValueError("invalid parameter, saturation <= 0")

    fowlerdata = numpy.asarray(fowlerdata)

    if fowlerdata.ndim != 3:
        raise ValueError('fowlerdata must be 3D')

    npairs = fowlerdata.shape[0] // 2
    if 2 * npairs != fowlerdata.shape[0]:
        raise ValueError('axis-0 in fowlerdata must be even')

    # change byteorder
    ndtype = fowlerdata.dtype.newbyteorder('=')
    fowlerdata = numpy.asarray(fowlerdata, dtype=ndtype)
    # type of the output
    fdtype = numpy.result_type(fowlerdata.dtype, dtype)
    # Type of the mask
    mdtype = numpy.dtype('uint8')

    fshape = (fowlerdata.shape[1], fowlerdata.shape[2])

    if badpixels is None:
        badpixels = numpy.zeros(fshape, dtype=mdtype)
    else:
        if badpixels.shape != fshape:
            raise ValueError('shape of badpixels is not '
                             'compatible with shape of fowlerdata')
        if badpixels.dtype != mdtype:
            raise ValueError('dtype of badpixels must be uint8')

    result = numpy.empty(fshape, dtype=fdtype)
    var = numpy.empty_like(result)
    npix = numpy.empty(fshape, dtype=mdtype)
    mask = badpixels.copy()

    _nirproc._process_fowler_intl(
        fowlerdata, ti, ts,  gain, ron,
        badpixels, saturation, blank,
        result, var, npix, mask
    )
    return result, var, npix, mask


def ramp_array(rampdata, ti, gain=1.0, ron=1.0,
               badpixels=None, dtype='float64',
               saturation=65631, blank=0, nsig=None, normalize=False):
    """Loop over the first axis applying ramp processing.

    *rampdata* is assumed to be a 3D numpy.ndarray containing the
    result of a nIR observation in folow-up-the-ramp mode.
    The shape of the array must be of the form N_s x M x N, with N_s being
    the number of samples.

    :param fowlerdata: Convertible to a 3D numpy.ndarray
    :param ti: Integration time.
    :param gain: Detector gain.
    :param ron: Detector readout noise in counts.
    :param badpixels: An optional MxN mask of dtype 'uint8'.
    :param dtype: The dtype of the float outputs.
    :param saturation: The saturation level of the detector.
    :param blank: Invalid values in output are substituted by *blank*.
    :returns: A tuple of signal, variance of the signal, numper of pixels used
        and badpixel mask.
    :raises: ValueError
    """

    import numina.array._nirproc as _nirproc
    if ti <= 0:
        raise ValueError("invalid parameter, ti <= 0.0")

    if gain <= 0:
        raise ValueError("invalid parameter, gain <= 0.0")

    if ron <= 0:
        raise ValueError("invalid parameter, ron < 0.0")

    if saturation <= 0:
        raise ValueError("invalid parameter, saturation <= 0")

    rampdata = numpy.asarray(rampdata)
    if rampdata.ndim != 3:
        raise ValueError('rampdata must be 3D')

    # change byteorder
    ndtype = rampdata.dtype.newbyteorder('=')
    rampdata = numpy.asarray(rampdata, dtype=ndtype)
    # type of the output
    fdtype = numpy.result_type(rampdata.dtype, dtype)
    # Type of the mask
    mdtype = numpy.dtype('uint8')
    fshape = (rampdata.shape[1], rampdata.shape[2])

    if badpixels is None:
        badpixels = numpy.zeros(fshape, dtype=mdtype)
    else:
        if badpixels.shape != fshape:
            msg = 'shape of badpixels is not compatible with shape of rampdata'
            raise ValueError(msg)
        if badpixels.dtype != mdtype:
            raise ValueError('dtype of badpixels must be uint8')

    result = numpy.empty(fshape, dtype=fdtype)
    var = numpy.empty_like(result)
    npix = numpy.empty(fshape, dtype=mdtype)
    mask = badpixels.copy()

    _nirproc._process_ramp_intl(
        rampdata, ti, gain, ron, badpixels,
        saturation, blank, result, var, npix, mask
    )
    return result, var, npix, mask


# This is not used...
# Old code used to detect cosmic rays in the ramp
def _ramp(data, saturation, dt, gain, ron, nsig):
    nsdata = data[data < saturation]

    # Finding glitches in the pixels
    intervals, glitches = _rglitches(nsdata, gain=gain, ron=ron, nsig=nsig)
    vals = numpy.asarray([_slope(nsdata[intls], dt=dt, gain=gain, ron=ron)
                          for intls in intervals if len(nsdata[intls]) >= 2])
    weights = (1.0 / vals[:, 1])
    average = numpy.average(vals[:, 0], weights=weights)
    variance = 1.0 / weights.sum()
    return average, variance, vals[:, 2].sum(), glitches


def _rglitches(nsdata, gain, ron, nsig):
    diffs = nsdata[1:] - nsdata[:-1]
    psmedian = numpy.median(diffs)
    sigma = math.sqrt(abs(psmedian / gain) + 2 * ron * ron)

    start = 0
    intervals = []
    glitches = []
    for idx, diff in enumerate(diffs):
        if not (psmedian - nsig * sigma < diff < psmedian + nsig * sigma):
            intervals.append(slice(start, idx + 1))
            start = idx + 1
            glitches.append(start)

    intervals.append(slice(start, None))

    return intervals, glitches


def _slope(nsdata, dt, gain, ron):

    if len(nsdata) < 2:
        raise ValueError('Two points needed to compute the slope')

    nn = len(nsdata)
    delt = dt * nn * (nn + 1) * (nn - 1) / 12
    ww = numpy.arange(1, nn + 1) - (nn + 1) / 2

    final = (ww * nsdata).sum() / delt

    # Readout limited case
    delt2 = dt * delt
    var1 = (ron / gain)**2 / delt2
    # Photon limiting case
    var2 = (6 * final * (nn * nn + 1)) / (5 * nn * dt * (nn * nn - 1) * gain)
    variance = var1 + var2
    return final, variance, nn



import pytest

from numina.tests.seffect import record_call

from ..wheel import Wheel, Carrousel


class Disp:
    def __init__(self, name):
        self.name = name


@pytest.fixture
def wheel_dev():
    wheel = Wheel("wheel", 3)
    for idx in range(3):
        wheel.put_in_pos(Disp(idx), idx)
    return wheel


@pytest.fixture
def carrousel_dev():
    wheel = Carrousel("wheel", 3)
    for idx in range(3):
        wheel.put_in_pos(Disp(idx), idx)
    return wheel


def test_wheel(wheel_dev):
    curr = wheel_dev.current()
    assert isinstance(curr, Disp)
    assert curr.name == 0
    assert wheel_dev.pos() == 0


def test_carrousel(carrousel_dev):
    curr = carrousel_dev.current()
    assert isinstance(curr, Disp)
    assert curr.name == 0
    assert carrousel_dev.pos() == 0


@pytest.mark.parametrize("pos", [0, 1, 2])
def test_move_wheel(carrousel_dev, pos):
    carrousel_dev.move_to(pos)

    curr = carrousel_dev.current()
    assert isinstance(curr, Disp)
    assert curr.name == pos
    assert carrousel_dev.pos() == pos


def test_move_wheel3(carrousel_dev):

    with pytest.raises(ValueError):
        carrousel_dev.move_to(8)


@pytest.mark.parametrize("pos", [1, 2])
def test_move_signals(carrousel_dev, pos):

    @record_call
    def changed_cb(pos):
        pass

    @record_call
    def moved_cb(pos):
        pass

    # Connect signals
    carrousel_dev.changed.connect(changed_cb)
    carrousel_dev.moved.connect(moved_cb)

    carrousel_dev.move_to(0)
    # Reset capture
    changed_cb.side_effect.clear()
    moved_cb.side_effect.clear()

    carrousel_dev.move_to(0)

    # Check call values
    assert moved_cb.side_effect.args == (0,)
    assert changed_cb.side_effect.called is False

    # Clear last call values
    changed_cb.side_effect.clear()
    moved_cb.side_effect.clear()

    carrousel_dev.move_to(pos)

    assert moved_cb.side_effect.called
    assert moved_cb.side_effect.args == (pos,)

    assert changed_cb.side_effect.called
    assert changed_cb.side_effect.args == (pos,)


@pytest.mark.parametrize("pos", [0, 1, 2])
def test_select(carrousel_dev, pos):
    carrousel_dev.select(pos)
    assert carrousel_dev.pos() == pos
    curr = carrousel_dev.current()
    assert isinstance(curr, Disp)
    assert curr.name == pos


@pytest.mark.parametrize("pos", [0, 1, 2])
def test_turn(wheel_dev, pos):
    wheel_dev.select(pos)

    @record_call
    def changed_cb(pos):
        pass

    @record_call
    def moved_cb(pos):
        pass

    # Connect signals
    wheel_dev.changed.connect(changed_cb)
    wheel_dev.moved.connect(moved_cb)

    steps = 6
    step = 0
    while step < steps:
        wheel_dev.turn()
        target = (pos + 1 + step) % 3
        step += 1
        assert wheel_dev.pos() == target
        assert moved_cb.side_effect.called
        assert moved_cb.side_effect.args == (target,)
        assert changed_cb.side_effect.called
        assert changed_cb.side_effect.args == (target,)

        changed_cb.side_effect.clear()
        moved_cb.side_effect.clear()


@pytest.mark.parametrize("pos", [1, 2])
def test_move_signals2(carrousel_dev, pos):

    @record_call
    def changed_cb(pos):
        pass

    @record_call
    def moved_cb(pos):
        pass

    # Connect signals
    carrousel_dev.changed.connect(changed_cb)
    carrousel_dev.moved.connect(moved_cb)

    carrousel_dev.move_to(0)
    # Reset capture
    changed_cb.side_effect.clear()
    moved_cb.side_effect.clear()

    carrousel_dev.move_to(0)

    # Check call values
    assert moved_cb.side_effect.called
    assert moved_cb.side_effect.args == (0,)
    assert changed_cb.side_effect.called is False
    # assert moved_cb.callpos == 0

    # Clear last call values
    changed_cb.side_effect.clear()
    moved_cb.side_effect.clear()

    carrousel_dev.move_to(pos)

    assert moved_cb.side_effect.called
    assert moved_cb.side_effect.args == (pos,)

    assert changed_cb.side_effect.called
    assert changed_cb.side_effect.args == (pos,)



#
# Copyright 2016-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


import numpy


class Stop(object):
    def __init__(self, name):
        self.name = name

    def transmission(self, wl):
        return numpy.zeros_like(wl)


class Open(object):
    def __init__(self, name):
        self.name = name

    def transmission(self, wl):
        return numpy.ones_like(wl)


class Filter(object):
    def __init__(self, name, transmission=None):
        self.name = name

    def transmission(self, wl):
        # FIXME: implement this with a proper
        # transmission
        return numpy.ones_like(wl)



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Interactive examination of 3D data cubes with ds9.
"""

from pathlib import Path
import shutil
import subprocess
import sys

import argparse
from astropy.io import fits
from astropy.wcs import WCS
import matplotlib.pyplot as plt
import numpy as np

from .ctext import ctext
from .extract_2d_slice_from_3d_cube import extract_slice


def ds9cmd(cmd, pipe=False):
    """Run a command in ds9 using xpa.

    Parameters
    ----------
    cmd : str
        The command to run in ds9.
    pipe : bool, optional
        If True, the command is run in a shell. This is useful
        when 'cm' contains a pipe, which gives an error when
        trying to run the command as a list.
        If False, the command is run as a list.
        In both cases, the output is captured.

    Returns
    -------
    str
        The output of the command, if successful.

    Raises
    ------
    ValueError
        If the command fails.
    """

    if pipe:
        result = subprocess.run(cmd, shell=True, capture_output=True, text=True, check=False)
    else:
        result = subprocess.run(cmd.split(), capture_output=True, text=True, check=False)

    if result.stderr == '':
        return result.stdout.strip()
    else:
        raise ValueError(result.stderr)


def init_ds9_plot(fpath, wave):
    """Initialize ds9 line plot."""
    ds9cmd('xpaset -p ds9 plot line {' +
           fpath.name +
           '} {' +
           'Value along NAXIS3 direction' +
           '} {' +
           'Signal' +
           '} xy')
    ds9cmd('xpaset -p ds9 plot font title size 12')
    ds9cmd('xpaset -p ds9 plot font labels size 12')
    ds9cmd('xpaset -p ds9 plot legend yes')
    ds9cmd(f'xpaset -p ds9 plot axis x min {np.min(wave.value)}')
    ds9cmd(f'xpaset -p ds9 plot axis x max {np.max(wave.value)}')
    ds9cmd('xpaset -p ds9 plot axis y min 0')
    ds9cmd('xpaset -p ds9 plot axis y max 1')
    ds9cmd('xpaset -p ds9 plot legend position top')
    current_plot = ds9cmd('xpaget ds9 plot current')
    return current_plot


def update_splot(data, source_mask, continuum_mask, wave,
                 fig, ax, line_objects, firstplot, plot_render):
    """Update plot with source and continuum spectra.

    Parameters
    ----------
    data : np.ndarray
        Array containing the 3D data cube.
    source_mask : np.ndarray
        The source mask.
    continuum_mask : np.ndarray
        The continuum mask.
    wave : np.ndarray
        Array with values along NAXIS3.
    fig : matplotlib.figure.Figure
        The Figure object containing the plot.
    ax : matplotlib.axes.Axes
        The Axes object of the plot.
    line_objects : list
        List of ax.plot() instances corresponding to the source,
        continuum and subtrated spectra.
    firstplot : bool
        If True, the next plot is the first one
    plot_render : str
        Display to display spectra: matplotlib or ds9
    """

    naxis3, naxis2, naxis1 = data.shape
    nmax = naxis2 * naxis1
    array2d_sp_source = np.zeros((nmax, naxis3))
    array2d_sp_continuum = np.zeros((nmax, naxis3))
    k_source = 0
    k_continuum = 0
    for i in range(naxis2):
        for j in range(naxis1):
            if source_mask[i, j] == 1:
                array2d_sp_source[k_source, :] = data[:, i, j]
                k_source += 1
            if continuum_mask[i, j] == 1:
                array2d_sp_continuum[k_continuum, :] = data[:, i, j]
                k_continuum += 1

    if k_source > 0:
        sp_source = np.nanmean(array2d_sp_source[:k_source, :], axis=0)
    else:
        sp_source = np.zeros(naxis3)
    sp_source_nonan = np.nan_to_num(sp_source, nan=0.0)

    if k_continuum > 0:
        sp_continuum = np.nanmean(array2d_sp_continuum[:k_continuum, :], axis=0)
    else:
        sp_continuum = np.zeros(naxis3)
    sp_continuum_nonan = np.nan_to_num(sp_continuum, nan=0.0)

    sp_subtracted = sp_source - sp_continuum
    sp_subtracted_nonan = sp_source_nonan - sp_continuum_nonan

    if plot_render in ['ds9', 'both']:
        if not firstplot:
            while ds9cmd('xpaget ds9 plot current dataset'):
                ds9cmd('xpaset -p ds9 plot delete dataset')

        for sp, sptype, npix, color in zip(
            [sp_subtracted_nonan, sp_source_nonan, sp_continuum_nonan],
            ['subtracted', 'source', 'continuum'],
            [None, k_source, k_continuum],
            ['blue', 'orange', 'green']
        ):
            np.savetxt(
                f'tmp_spectrum_{sptype}.dat',
                np.column_stack((wave.value, sp)),
                fmt='%e'
            )
            cnpix = f' ({npix})' if npix is not None else ''
            ds9cmd(f'cat tmp_spectrum_{sptype}.dat | xpaset ds9 plot data xy', pipe=True)
            ds9cmd(f'xpaset -p ds9 plot line color {color}')
            ds9cmd('xpaset -p ds9 plot name {' +
                   f'{sptype}' + cnpix +
                   '}')

    if plot_render in ['matplotlib', 'both']:
        # recompute global Y-axis limits
        sp_concatenate = np.concatenate((sp_source, sp_continuum, sp_subtracted))
        ymin = np.nanmin(sp_concatenate)
        ymax = np.nanmax(sp_concatenate)
        dy = ymax - ymin
        if dy == 0:
            dy = 0.1
        ymin -= dy / 20
        ymax += dy / 20
        ax.set_ylim(ymin, ymax)

        line_source, line_continuum, line_subtracted = line_objects
        line_source.set_ydata(sp_source)
        line_source.set_label(f'source ({k_source})')
        line_continuum.set_ydata(sp_continuum)
        line_continuum.set_label(f'continuum ({k_continuum})')
        line_subtracted.set_ydata(sp_subtracted)
        ax.draw_artist(line_source)
        ax.draw_artist(line_continuum)
        ax.draw_artist(line_subtracted)
        # The legend in Matplotlib is a separate artist that doesn't automatically
        # udpate when you change plot elements unless you explicitly update or redraw it.
        # For that reason it is necessary to remove the old legend and redraw the new one.
        old_legend = ax.get_legend()
        if old_legend:
            old_legend.remove()
        ax.legend(loc='upper center', bbox_to_anchor=(0.5, 1.15), ncol=3)
        fig.canvas.draw()
        fig.canvas.flush_events()


def update_ds9regions(data, source_mask, continuum_mask, tmp_mask, wave,
                      fig, ax, line_objects, firstplot, plot_render):
    """Update ds9 regions interactively using the source and continuum masks.

    Parameters
    ----------
    data : np.array
        Array containing the 3D data cube.
    source_mask : np.ndarray
        The source mask to update.
    continuum_mask : np.ndarray
        The continuum mask to update.
    tmp_mask : np.ndarray or None
        A temporary mask to display starting point of a new
        rectangular region.
    wave : np.ndarray
        Array with values along NAXIS3.
    fig : matplotlib.figure.Figure
        The Figure object containing the plot.
    ax : matplotlib.axes.Axes
        The Axes object of the plot.
    line_objects: list
        List of ax.plot() instances corresponding to the source,
        continuum and subtrated spectra.
    firstplot = bool
        If True, the next plot is the first one.
    plot_render : str
        Display to display spectra: matplotlib or ds9
    """
    # update file with regions
    lines = [
        '# Region file format: DS9 version 4.1',
        'global color=green dashlist=8 3 width=1 font="helvetica 10 normal roman"' +
        'select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source=1',
        'physical'
    ]
    naxis2, naxis1 = source_mask.shape
    for i in range(naxis2):
        for j in range(naxis1):
            if source_mask[i, j]:
                lines.append(f'box({j+1},{i+1},1,1,0) # fill=1 color=red source=1')
            if continuum_mask[i, j]:
                lines.append(f'box({j+1},{i+1},1,1,0) # fill=1 color=green source=0')
            if tmp_mask is not None:
                if tmp_mask[i, j] != 0:
                    lines.append(f'box({j+1},{i+1},1,1,0) # fill=1 color=cyan')
    with open('tmp_regions_ds9.reg', 'wt', encoding='ascii') as f:
        for line in lines:
            f.write(line + '\n')
    try:
        ds9cmd('xpaset -p ds9 region delete')
        ds9cmd('xpaset -p ds9 region load tmp_regions_ds9.reg')
    except ValueError as exc:
        print(f'WARNING: {exc}')
    # Update splot
    update_splot(
        data=data,
        source_mask=source_mask,
        continuum_mask=continuum_mask,
        wave=wave,
        fig=fig,
        ax=ax,
        line_objects=line_objects,
        firstplot=firstplot,
        plot_render=plot_render
    )


def display_help_menu(plot_render):
    """Display help text when selecting pixels.

    Parameters
    ----------
    plot_render: str
        Display to display spectra: matplotlib or ds9
    """
    print("Click on the ds9 window to select pixels:")
    print("  - Press 's' to select a single source pixel")
    print("  - Press 'c' to select a single continuum pixel")
    print("  - Press 'x' to remove a single pixel from any mask")
    print("  - Press 'r' to reset both masks")
    print("  - Press 'a' to start selecting a rectangular region")
    print("    (then press 's' or 'c' in the opposite corner to define the mask type)")
    if plot_render in ['matplotlib', 'both']:
        print("  - Press 'p' to pause pixel selection and allow matplotlib interaction")
    print("  - Press 'q' to quit (stop pixel selection)")
    print("  - Press 'h' to display this help")


def update_masks(filename, data, source_mask, continuum_mask, wave, verbose, plot_render):
    """Update source and continuum masks interactively using ds9.

    Parameters
    ----------
    filename : str
        File name retrieved from ds9.
    data : np.ndarray
        Array containing the 3D data cube.
    source_mask : np.ndarray
        The source mask to update.
    continuum_mask : np.ndarray
        The continuum mask to update.
    wave : astropy.units.Quantity
        The wavelength array corresponding to the spectral axis.
    verbose : bool
        If True, print verbose output.
    plot_render : str
        Display to display spectra: matplotlib or ds9

    """

    if source_mask.shape != continuum_mask.shape:
        print(f'{source_mask.shape=}')
        print(f'{continuum_mask.shape=}')
        raise ValueError('Incompatible mask shapes')
    naxis2, naxis1 = source_mask.shape
    tmp_mask = None

    if plot_render in ['matplotlib', 'both']:
        plt.ion()  # enable interactive mode
        fig, ax = plt.subplots()
        sp_source = np.zeros(len(wave))
        sp_continuum = np.zeros(len(wave))
        sp_subtracted = np.zeros(len(wave))
        (line_subtracted,) = ax.plot(wave, sp_subtracted, 'C0-', label='subtracted', zorder=3)
        (line_source,) = ax.plot(wave, sp_source, 'C1-', label='source', zorder=2)
        (line_continuum,) = ax.plot(wave, sp_continuum, 'C2-', label='continuum', zorder=1)
        ax.legend(loc='upper center', bbox_to_anchor=(0.5, 1.15), ncol=3)
        line_objects = [line_source, line_continuum, line_subtracted]
        ax.set_title(f'{filename}')
        ax.set_xlabel('Value along NAXIS3 direction')
        ax.set_ylabel('Signal')
    else:
        fig = None
        ax = None
        line_objects = [None, None, None]

    update_ds9regions(
        data=data,
        source_mask=source_mask,
        continuum_mask=continuum_mask,
        tmp_mask=tmp_mask,
        wave=wave,
        fig=fig,
        ax=ax,
        line_objects=line_objects,
        firstplot=True,
        plot_render=plot_render
    )

    display_help_menu(plot_render)

    loop = True
    last_key_pos = [None, None, None]
    ix1, ix2, iy1, iy2 = None, None, None, None  # avoid PyCharm warning
    while loop:
        try:
            key, x, y = ds9cmd('xpaget ds9 iexam key coordinate image').split()
        except ValueError as exc:
            if verbose:
                print(f'WARNING: {exc}')
        if key in ['s', 'c', 'r', 'a', 'x']:
            x = str(round(float(x)))
            y = str(round(float(y)))
            if verbose:
                print(f'key: {key}: selecting pixel {x=}, {y=}')
            iy = int(y) - 1
            ix = int(x) - 1
            if key == 'a':
                last_key_pos = [key, ix, iy]
                tmp_mask = np.zeros(shape=(naxis2, naxis1), dtype=np.uint8)
                tmp_mask[iy, ix] = 1
            elif key in ['s', 'c'] and last_key_pos[0] is not None:
                ix1 = min(ix, last_key_pos[1])
                ix2 = max(ix, last_key_pos[1])
                iy1 = min(iy, last_key_pos[2])
                iy2 = max(iy, last_key_pos[2])
                last_key_pos = [None, None, None]
            elif key in ['s', 'c', 'x']:
                ix1 = ix
                ix2 = ix
                iy1 = iy
                iy2 = iy
                # last_key_pos = [key, ix, iy]
            if key in ['s', 'c', 'x']:
                for iy in range(iy1, iy2+1):
                    for ix in range(ix1, ix2+1):
                        if key == 's':
                            if continuum_mask[iy, ix] == 0:
                                source_mask[iy, ix] = 1
                            else:
                                continuum_mask[iy, ix] = 0
                                source_mask[iy, ix] = 1
                        elif key == 'c':
                            if source_mask[iy, ix] == 0:
                                continuum_mask[iy, ix] = 1
                            else:
                                source_mask[iy, ix] = 0
                                continuum_mask[iy, ix] = 1
                        elif key == 'x':
                            source_mask[iy, ix] = 0
                            continuum_mask[iy, ix] = 0
                        else:
                            raise ValueError('Unexpected error')
            elif key == 'r':
                for i in range(naxis2):
                    for j in range(naxis1):
                        source_mask[i, j] = 0
                        continuum_mask[i, j] = 0
            elif key == 'a':
                pass
            else:
                raise ValueError('Unexpected error')
            # update file with regions
            update_ds9regions(
                data=data,
                source_mask=source_mask,
                continuum_mask=continuum_mask,
                tmp_mask=tmp_mask,
                wave=wave,
                fig=fig,
                ax=ax,
                line_objects=line_objects,
                firstplot=False,
                plot_render=plot_render
            )
            if key == 'a':
                tmp_mask = None
        elif key == 'p':
            if plot_render in ['matplotlib', 'both']:
                input('Press RETURN to continue with pixel selection...')
        elif key == 'h':
            display_help_menu(plot_render)
        elif key == 'q':
            cquit = input('Do you want to quit? (y/[n]) ')
            if cquit.lower() in ['y', 'yes']:
                loop = False
                if verbose:
                    print('Selection of pixels finished!')

    if plot_render in ['matplotlib', 'both']:
        # keep the splot open after updates
        plt.ioff()
        print("Press 'q' to close matplotlib window and stop the program")
        plt.show(block=True)


def main(args=None):
    """Main function"""
    parser = argparse.ArgumentParser(
        description="Interactive examination of 3D data cubes with ds9."
    )

    parser.add_argument("datacube",
                        help="Input 3D FITS data cube",
                        type=str)
    parser.add_argument("--i1",
                        help='First pixel along NAXIS3 (default 1)',
                        type=int, default=1)
    parser.add_argument("--i2",
                        help='Last pixel along NAXIS3 (default NAXIS3)',
                        type=str)
    parser.add_argument("--ds9exec",
                        help="Command line to launch ds9 (default 'ds9')",
                        type=str, default='ds9')
    parser.add_argument("--plot_render",
                        help="Display to display spectra (default=matplotlib)",
                        choices=['matplotlib', 'ds9', 'both'],
                        default='matplotlib')
    parser.add_argument("--input_masks",
                        help="Path to a FITS file with source and continuum masks",
                        type=str)
    parser.add_argument("--output_masks",
                        help="Path to the output FITS file with source and continuum masks",
                        type=str)
    parser.add_argument("--verbose",
                        help="Display intermediate information",
                        action="store_true")
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if args.verbose:
        for arg, value in vars(args).items():
            print(ctext(f'{arg}: {value}', faint=True))

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    file_datacube = args.datacube
    ds9exec = args.ds9exec
    verbose = args.verbose
    plot_render = args.plot_render

    # Check XPA is installed
    list_required_executables = ['xpaget', 'xpaset']
    for required_executable in list_required_executables:
        execfile = shutil.which(required_executable)
        if execfile is None:
            raise SystemExit(f'The program {required_executable} is not available')
        else:
            if verbose:
                print(f'Required program {execfile} found!')

    # Check if ds9 is already running
    print('Checking if a previous ds9 instance is running...')
    try:
        filename = ds9cmd('xpaget ds9 file')
    except Exception as exc:
        filename = ''
        if verbose:
            print(f'{exc=}')
            print('No previous ds9 instance found. OK!')
    if len(filename) > 0:
        print('A previous instance of ds9 is already running.\n' +
              f'Filename: {filename}\n' +
              'Please close it before using this program')
        raise SystemExit()

    # Get header and data of the FITS file
    fpath = Path(file_datacube)
    header = fits.getheader(fpath)
    wcs = WCS(header)
    if verbose:
        print(f'WCS: {wcs}')

    data = fits.getdata(fpath)
    if len(data.shape) != 3:
        raise ValueError(f'Expected a 3D cube, but got {data.shape}')

    naxis3, naxis2, naxis1 = data.shape
    if verbose:
        print(f'{naxis1=}')
        print(f'{naxis2=}')
        print(f'{naxis3=}')

    i1 = args.i1
    if i1 < 1 or i1 > naxis3:
        raise ValueError(f'Invalid first pixel={i1} along NAXIS3={naxis3}')
    if args.i2 is None:
        i2 = naxis3
    else:
        i2 = int(args.i2)
        if i2 < i1 or i2 > naxis3:
            raise ValueError(f'Invalid last pixel={i2} along NAXIS3={naxis3}')

    # Collapse the data cube along NAXIS3
    if verbose:
        print('Collapsing 3D cubes along NAXIS3... ', end='')
    extract_slice(
        input=file_datacube,
        axis=3,
        i1=i1,
        i2=i2,
        method='sum',
        wavecal='none',
        transpose=False,
        vmin=None,
        vmax=None,
        noplot=True,
        output='tmp_collapsed_3D.fits'
    )
    print('OK!')

    # Launch ds9
    cmd = f"{ds9exec.split()[0]} tmp_collapsed_3D.fits {' '.join(ds9exec.split()[1:])} &"
    if verbose:
        print('Executing:')
        print(cmd)
    # Note: use shell=True below to make the ds9 alias in the system available
    result = subprocess.run(cmd, capture_output=True, text=True, check=False, shell=True)
    if verbose:
        print(f'{result.stderr=}')
        print(f'{result.stdout=}')
    input('Press RETURN after ds9 has properly started...')
    try:
        filename = ds9cmd('xpaget ds9 file')
    except Exception as exc:
        raise SystemExit('Fatal error: ds9 is not running') from exc
    if verbose:
        print(f'ds9 working with file: {filename}')

    # Generate array in the spectral direction
    wcs1d_spectral = wcs.spectral
    wave = wcs1d_spectral.pixel_to_world(np.arange(naxis3))
    if verbose:
        print(f'Minimum value along NAXIS3: {wave.min()}')
        print(f'Maximum value along NAXIS3: {wave.max()}')

    # Read source and continuum masks or create them
    if args.input_masks:
        source_mask = fits.getdata(args.input_masks, extname='SOURMASK')
        if source_mask.shape != (naxis2, naxis1):
            raise ValueError(f'{source_mask.shape=} does not match {(naxis2, naxis1)=}')
        if source_mask.dtype != np.uint8:
            raise ValueError(f'Source mask dtype {source_mask.dtype} is not uint8')
        continuum_mask = fits.getdata(args.input_masks, extname='CONTMASK')
        if continuum_mask.shape != (naxis2, naxis1):
            raise ValueError(f'{continuum_mask.shape=} does not match {(naxis2, naxis1)=}')
        if continuum_mask.dtype != np.uint8:
            raise ValueError(f'Continuum mask dtype {continuum_mask.dtype} is not uint8')
    else:
        source_mask = np.zeros((naxis2, naxis1), dtype=np.uint8)
        continuum_mask = np.zeros((naxis2, naxis1), dtype=np.uint8)

    if plot_render in ['ds9', 'both']:
        current_plot = init_ds9_plot(fpath=fpath, wave=wave)
        if verbose:
            print(f'Opening ds9 plot: {current_plot}')

    update_masks(
        filename=file_datacube,
        data=data,
        source_mask=source_mask,
        continuum_mask=continuum_mask,
        wave=wave,
        verbose=verbose,
        plot_render=plot_render
    )

    # Save the masks to a FITS file
    if args.output_masks:
        output_masks = Path(args.output_masks)
    else:
        output_masks = Path('tmp_masks.fits')
    hdu0 = fits.PrimaryHDU()
    hdu1 = fits.ImageHDU(source_mask, name='SOURMASK')
    hdu2 = fits.ImageHDU(continuum_mask, name='CONTMASK')
    hdul = fits.HDUList([hdu0, hdu1, hdu2])
    hdul.writeto(output_masks, overwrite=True)
    if verbose:
        print(f'Masks saved to {output_masks}')

    print('Remember to close the running session of ds9 before re-executing this program!')


if __name__ == "__main__":
    main()




from ..load import load


def test_load_base():

    class A(object):
        pass

    tag = A()
    obj = 0

    assert obj == load(tag, obj)


def test_load_method():

    class A(object):

        def _datatype_load(self, obj):
            return obj + 1

    tag = A()
    obj = 1

    assert obj + 1 == load(tag, obj)


def test_load_method_deprecated():

    class B(object):

        def __numina_load__(self, obj):
            return obj + 2

    tag = B()
    obj = 1

    assert obj + 2 == load(tag, obj)


# @pytest.mark.skipif(sys.version_info < (3, 4),
#                     reason="https://github.com/pytest-dev/pytest/issues/840")
# def test_dump_method_deprecated_warning(recwarn):
#     warnings.simplefilter('always')
#     test_load_method_deprecated()
#
#     assert recwarn.pop(DeprecationWarning)


def test_load_method_register():

    class C(object):
        pass

    def numina_load_func(tag, obj):
        return obj + 3

    load.register(C, numina_load_func)

    tag = C()
    obj = 1

    assert obj + 3 == load(tag, obj)





import os
from ..context import environ


def test_environ_context1():
    """Remove env var if it was not set"""

    cache_dir = '/cache/dir'

    with environ(OOO_CACHE_HOME=cache_dir):
        assert os.environ['OOO_CACHE_HOME'] == cache_dir

    is_in = 'OOO_CACHE_HOME' in os.environ

    assert is_in is False


def test_environ_context2():
    """Reset env var if it was set"""

    cache_dir1 = '/cache/dir/1'
    cache_dir2 = '/cache/dir/2'

    os.environ['OOO_CACHE_HOME'] = cache_dir1

    with environ(OOO_CACHE_HOME=cache_dir2):
        assert os.environ['OOO_CACHE_HOME'] == cache_dir2

    assert os.environ['OOO_CACHE_HOME'] == cache_dir1


def test_environ_context3():
    """Reset multiple variables"""

    cache_dir1a = '/cache/dir/1a'
    cache_dir2a = '/cache/dir/2a'
    cache_dir3a = '/cache/dir/3a'
    cache_dir1b = '/cache/dir/1b'
    cache_dir2b = '/cache/dir/2b'

    os.environ['OOO_CACHE_HOME1'] = cache_dir1a
    os.environ['OOO_CACHE_HOME2'] = cache_dir2a
    os.environ['OOO_CACHE_HOME3'] = cache_dir3a

    with environ(OOO_CACHE_HOME1=cache_dir1b,
                 OOO_CACHE_HOME2=cache_dir2b):
        assert os.environ['OOO_CACHE_HOME1'] == cache_dir1b
        assert os.environ['OOO_CACHE_HOME2'] == cache_dir2b
        assert os.environ['OOO_CACHE_HOME3'] == cache_dir3a

    assert os.environ['OOO_CACHE_HOME1'] == cache_dir1a
    assert os.environ['OOO_CACHE_HOME2'] == cache_dir2a
    assert os.environ['OOO_CACHE_HOME3'] == cache_dir3a



#
# Copyright 2008-2014 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import unittest

import numpy

import numina.array as array


class ArrayTestCase(unittest.TestCase):

    def test_subarray_match(self):
        '''Test subarray_match.'''

        # Shapes don't intersect
        minor, major = array.subarray_match((3, 4, 5),
                                            (10, 10, 10),
                                            (2, 2, 2))

        #  Returns None if shapes don't intersect
        self.assertTrue(minor is None,)
        self.assertTrue(major is None)

        small = 100

        # One is contained inside the other
        minor, major = array.subarray_match((2048, 2048),
                                            (0, 0), (small, small))
        # If one array is inside the other
        # the outputs is equal to the small array
        self.assertEqual(minor, (slice(0, small, None),
                                 slice(0, small, None)))
        self.assertEqual(major, (slice(0, small, None),
                                 slice(0, small, None)))

        # One is contained inside the other, with offsets
        minor, major = array.subarray_match((2048, 2048), (30, 40),
                                            (small, small))
        # If one array is inside the other
        # the outputs is equal to the small array
        self.assertEqual(minor, (slice(30, small + 30, None),
                                 slice(40, small + 40, None)))
        self.assertEqual(major, (slice(0, small, None),
                                 slice(0, small, None)))

        # Both are equal, with offsets
        minor, major = array.subarray_match((100, 100),
                                            (30, 40), (small, small))
        # If one array is inside the other
        # the outputs is equal to the small array
        self.assertEqual(minor, (slice(30, small, None),
                                 slice(40, small, None)))
        self.assertEqual(major, (slice(0, small - 30, None),
                                 slice(0, small - 40, None)))

        # Equal offsets in both sides
        minor, major = array.subarray_match((100, 100), (30, 40),
                                            (100, 100), (30, 40))
        # If one array is inside the other
        # the outputs is equal to the small array
        self.assertEqual(minor, (slice(0, small, None),
                                 slice(0, small, None)))
        self.assertEqual(major, (slice(0, small, None),
                                 slice(0, small, None)))

        # Different offsets in both sides
        minor, major = array.subarray_match((100, 100), (31, 42),
                                            (100, 100), (10, 20))
        # If one array is inside the other
        # the outputs is equal to the small array
        self.assertEqual(minor, (slice(21, small, None),
                                 slice(22, small, None)))
        self.assertEqual(major, (slice(0, 79, None), slice(0, 78, None)))

        # If we interchange the arrays and change the sign of the offset,
        # we get the same result
        minor, major = array.subarray_match((100, 100), (10, 20),
                                            (200, 100))
        cminor, cmajor = array.subarray_match((200, 100), (-10, -20),
                                              (100, 100))
        self.assertEqual(cminor, major)
        self.assertEqual(cmajor, minor)


class FixpixTestCase(unittest.TestCase):
    def test_exception(self):
        data = numpy.zeros((10, 10))
        mask = numpy.zeros((100, 100))

        self.assertRaises(ValueError, array.fixpix, data, mask)

    def test_array_is_the_same(self):
        data = numpy.zeros((10, 10))
        mask = numpy.zeros((10, 10))

        data2 = array.fixpix(data, mask)

        self.assertIs(data2, data)

    def test_simple_interpolation(self):
        data = numpy.array([[1.0, -1000, 3.0], [4.0, -1000, 6.0],
                            [7.0, -10000, 9.0]])
        mask = numpy.array([[0, 1, 0], [0, 1, 0], [0, 1, 0]])

        data2 = array.fixpix(data, mask)
        for i, v in zip(data2[:, 1], [2.0, 5.0, 8.0]):
            self.assertAlmostEqual(i, v)



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from astropy.units import Unit
import numpy as np

from numina.array.distortion import fmap


def update_image2d_rss_detector_method0(
        islice,
        simulated_x_ifu_all,
        simulated_y_ifu_all,
        simulated_wave_all,
        naxis1_ifu,
        bins_x_ifu,
        bins_wave,
        bins_x_detector,
        bins_y_detector,
        wv_cdelt1,
        extra_degradation_spectral_direction,
        dict_ifu2detector,
        image2d_rss_method0,
        image2d_detector_method0
):
    """Update the two 2D images: RSS and detector.

    The function updates the following 2D arrays:
    - image2d_rss_method0,
    - image2d_detector_method0
    with the photons observed through the slice 'islice'.

    Note that both arrays are generated simultaneously in
    order to make use of the same value of
    'extra_degradation_spectral_direction'.

    This function can be executed in parallel.

    Parameters
    ----------
    islice : int
        Slice number.
    simulated_x_ifu_all : `~astropy.units.Quantity`
        Simulated X coordinates of the photons in the IFU.
    simulated_y_ifu_all : `~astropy.units.Quantity`
        Simulated Y coordinates of the photons in the IFU.
    simulated_wave_all : `~astropy.units.Quantity`
        Simulated wavelengths of the photons in the IFU.
    naxis1_ifu : `~astropy.units.Quantity`
        IFU NAXIS1 (along the slice)
    bins_x_ifu : `~numpy.ndarray`
        Bin edges in the naxis1_ifu direction
        (along the slice).
    bins_wave : `~numpy.ndarray`
        Bin edges in the wavelength direction.
    bins_x_detector : `~numpy.ndarray`
        Bin edges in the naxis1_detector direction
        (spectral direction).
    bins_y_detector : `~numpy.ndarray`
        Bin edges in the naxis2_detector direction
        (slices direction).
    wv_cdelt1 : `~astropy.units.Quantity`
        CDELT1 value along the spectral direction.
    extra_degradation_spectral_direction : `~astropy.units.Quantity`
        Additional degradation in the spectral direction, in units of
        the detector pixels, for each simulated photon.
    dict_ifu2detector : dict
        A Python dictionary containing the 2D polynomials that allow
        to transform (X, Y, wavelength) coordinates in the IFU focal
        plane to (X, Y) coordinates in the detector.
    image2d_rss_method0 : `~numpy.ndarray`
        2D array containing the RSS image. This array is
        updated by this function.
    image2d_detector_method0 : `~numpy.ndarray`
        2D array containing the detector image. This array is
        updated by this function.

    """

    # determine photons that pass through the considered slice
    y_ifu_expected = 1.5 + 2 * islice
    condition = np.abs(simulated_y_ifu_all.value - y_ifu_expected) < 1
    iok = np.where(condition)[0]
    nphotons_slice = len(iok)

    if nphotons_slice > 0:
        # -------------------------------------------------
        # 1) spectroscopic 2D image with continguous slices
        # -------------------------------------------------
        h, xedges, yedges = np.histogram2d(
            x=simulated_x_ifu_all.value[iok],
            y=simulated_wave_all.value[iok] +
              (simulated_y_ifu_all.value[iok] - y_ifu_expected) * wv_cdelt1.value +
              extra_degradation_spectral_direction.value[iok] * wv_cdelt1.value,
            bins=(bins_x_ifu.value, bins_wave.value)
        )
        j1 = islice * naxis1_ifu.value
        j2 = j1 + naxis1_ifu.value
        image2d_rss_method0[j1:j2, :] += h

        # -----------------------------------------
        # 2) spectroscopic 2D image in the detector
        # -----------------------------------------
        # use models to predict location in Hawaii detector
        # important: reverse here X <-> Y
        wavelength_unit = Unit(dict_ifu2detector['wavelength-unit'])
        dumdict = dict_ifu2detector['contents'][islice]
        order = dumdict['order']
        aij = np.array(dumdict['aij'])
        bij = np.array(dumdict['bij'])
        y_hawaii, x_hawaii = fmap(
            order=order,
            aij=aij,
            bij=bij,
            x=simulated_x_ifu_all.value[iok],
            # important: use the wavelength unit employed to determine
            # the polynomial transformation
            y=simulated_wave_all.to(wavelength_unit).value[iok]
        )
        # disperse photons along the spectral direction according to their
        # location within the slice in the vertical direction
        x_hawaii += simulated_y_ifu_all[iok].value - y_ifu_expected
        # include additional degradation in spectral resolution
        x_hawaii += extra_degradation_spectral_direction.value[iok]
        # compute 2D histogram
        # important: reverse X <-> Y
        h, xedges, yedges = np.histogram2d(
            x=y_hawaii,
            y=x_hawaii,
            bins=(bins_y_detector, bins_x_detector)
        )
        image2d_detector_method0 += h
